"""Tools for importing and exporting data."""
from __future__ import annotations

import io
import json
import pathlib
from itertools import chain
from typing import Iterable, List, Optional, Set, TextIO, Union

from rdflib import OWL, RDF, XSD, Graph, Literal, URIRef
from rdflib.parser import Parser as RDFLib_Parser
from rdflib.plugin import get as get_plugin
from rdflib.serializer import Serializer as RDFLib_Serializer
from rdflib.term import Identifier
from rdflib.util import guess_format

from simphony_osp.ontology import OntologyIndividual
from simphony_osp.session.session import Session
from simphony_osp.utils.simphony_namespace import simphony_namespace

__all__ = ["export_file", "import_file"]


def import_file(
    path_or_filelike: Union[str, TextIO, dict, List[dict]],
    session: Optional[Session] = None,
    format: str = None,
) -> Union[OntologyIndividual, Set[OntologyIndividual]]:
    """Imports ontology individuals from a file and load them into a session.

    Args:
        path_or_filelike: either,
            (str) the path of a file to import;
            (Union[List[dict], dict]) a dictionary representing the contents of
             a json file;
            (TextIO) any file-like object  (in string mode) that provides a
            `read()` method. Note that it is possible to get such an object
            from any `str` object using the python standard library. For
            example, given the `str` object `string`, `import io;
            filelike = io.StringIO(string)` would create such an object.
            If not format is specified, it will be guessed.
        session: the session in which the imported data will be
            stored.
        format: the format of the content to import. The supported formats
            are the ones supported by RDFLib. See
            `https://rdflib.readthedocs.io/en/latest/plugin_parsers.html`.
            If no format is specified, then it will be guessed. Note that in
            some specific cases, the guess may be wrong. In such cases, try
            again specifying the format.

    Returns:
        A set with the imported ontology individuals. If an individual was
        defined as the "main" one using the SimPhoNy ontology, then only the
        main individual is returned instead.
    """
    # Check that the requested format is supported and raise useful exceptions.
    if format is not None:
        try:
            get_plugin(format, RDFLib_Parser)
        except AttributeError as e:
            if "/" not in format:
                raise ValueError(
                    f"Unsupported format {format}. The supported formats are "
                    f"the ones supported by RDFLib "
                    f"`https://rdflib.readthedocs.io/en/latest"
                    f"/plugin_parsers.html`."
                ) from e
            else:
                raise ValueError(
                    f"Unsupported mime-type {format}. The supported mime-types"
                    f" are the ones supported by RDFLib. Unfortunately, the "
                    f"latter are not documented, but can be checked directly "
                    f"on its source code "
                    f"`https://github.com/RDFLib/rdflib/blob/master"
                    f"/rdflib/plugin.py`. Look for lines of the form "
                    f'`register(".*", Parser, ".*", ".*")`.'
                ) from e

    # Guess and/or validate the specified format.
    if isinstance(path_or_filelike, (dict, list)):  # JSON document
        if not (
            format is None or format in ("json-ld", "application/ld+json")
        ):
            raise ValueError(
                f"The file to be imported do not match the specified format: "
                f"{format}."
            )
        contents = json.dumps(path_or_filelike)
    else:  # Path to a file or file-like object.
        # Read the contents of the object.
        if isinstance(path_or_filelike, str):  # Path.
            if not pathlib.Path(path_or_filelike).is_file():
                raise ValueError(
                    f"{path_or_filelike} is not a file or does not exist."
                )
            with open(path_or_filelike, "r") as file:
                contents = file.read()
        else:  # File-like object.
            if "read" not in path_or_filelike.__dir__():
                raise TypeError(
                    f"{path_or_filelike} is neither a path"
                    f"or a file-like object."
                )
            contents = path_or_filelike.read()

        # Guess or validate the format.
        if format is None:
            # Let RDFLib guess (it can only guess for files)
            if isinstance(path_or_filelike, str):
                format = guess_format(path_or_filelike)
            else:
                raise ValueError(
                    "Could not guess the file format. Please"
                    'specify it using the "format" keyword '
                    "argument."
                )

    # Import the contents.
    session = session or Session.get_default_session()

    buffer_session = Session()
    buffer_session.graph.parse(io.StringIO(contents), format=format)
    buffer_session.graph.remove((None, RDF.type, OWL.NamedIndividual))

    individuals = set(
        individual
        for individual in buffer_session
        if isinstance(individual, OntologyIndividual)
    )
    session.add(individuals, exists_ok=True, merge=False)

    main = next(
        iter(
            buffer_session.graph.subjects(
                simphony_namespace.main, Literal("true", datatype=XSD.boolean)
            )
        ),
        None,
    )
    if main:
        result = session.from_identifier_typed(main, typing=OntologyIndividual)
    else:
        result = set(
            session.from_identifier_typed(
                individual.identifier, typing=OntologyIndividual
            )
            for individual in individuals
        )

    return result


def export_file(
    individuals_or_session: Optional[
        Union[OntologyIndividual, Iterable[OntologyIndividual], Session]
    ] = None,
    file: Optional[Union[str, TextIO]] = None,
    main: Optional[Union[str, Identifier, OntologyIndividual]] = None,
    format: str = "text/turtle",
) -> Union[str, None]:
    """Exports ontology individuals to a variety of formats.

    Args:
        individuals_or_session:
            (OntologyIndividual) A single ontology individual to export, or
            (Iterable[OntologyIndividual]) an iterable of ontology individuals,
            (Session) a session to serialize all of its ontology individuals.
            If `None` is specified, then the current session is exported.
        file: either,
            (str) a path, to save the ontology individuals to, or
            (TextIO) any file-like object (in string mode) that provides a
            `write()` method. If this argument is not specified, a string with
            the results will be returned instead.
        main: the identifier of an ontology individual to be marked as the
            "main" individual using the SimPhoNy ontology.
        format: the target format. Defaults to triples in turtle syntax.

    Returns:
        The contents of the exported file as a string (if no `file` argument
        was provided), or nothing.
    """
    # Choose default session if not specified.
    individuals_or_session = (
        individuals_or_session or Session.get_default_session()
    )

    # Check the validity of the requested format and raise useful exceptions.
    try:
        get_plugin(format, RDFLib_Serializer)
    except AttributeError as e:
        if "/" not in format:
            raise ValueError(
                f"Unsupported format {format}. The supported formats are "
                f"the ones supported by RDFLib "
                f"`https://rdflib.readthedocs.io/en/latest"
                f"/plugin_serializers.html`."
            ) from e
        else:
            raise ValueError(
                f"Unsupported mime-type {format}. The supported mime-types"
                f" are the ones supported by RDFLib. Unfortunately, the "
                f"latter are not documented, but can be checked directly "
                f"on its source code "
                f"`https://github.com/RDFLib/rdflib/blob/master"
                f"/rdflib/plugin.py`. Look for lines of the form "
                f'`register(".*", Parser, ".*", ".*")`.'
            ) from e

    # Compute main individual triple if applicable
    if isinstance(individuals_or_session, OntologyIndividual) and not main:
        main = individuals_or_session
    if main:
        if isinstance(main, str):
            main = URIRef(main)
        elif isinstance(main, OntologyIndividual):
            main = main.identifier
        main = (
            main,
            simphony_namespace.main,
            Literal("true", datatype=XSD.boolean),
        )

    # Export the individuals
    if isinstance(individuals_or_session, Session):
        buffer_graph = Graph()
        buffer_graph += individuals_or_session.graph
        if main:
            buffer_graph.add(main)
        for ns in individuals_or_session.namespaces:
            buffer_graph.bind(ns.name, ns.iri)
        result = buffer_graph.serialize(
            format=format, encoding="utf-8"
        ).decode("utf-8")
    else:
        if isinstance(individuals_or_session, OntologyIndividual):
            individuals_or_session = {individuals_or_session}

        buffer_session = Session()

        # Bind namespaces
        first_individual = next(iter(individuals_or_session))
        for ns in first_individual.session.namespaces:
            buffer_session.bind(ns.name, ns.iri)
        individuals_or_session = chain(
            {first_individual}, individuals_or_session
        )

        # Add individuals and main individual triple
        buffer_session.add(individuals_or_session)
        if main:
            buffer_session.graph.add(main)

        result = buffer_session.graph.serialize(
            format=format, encoding="utf-8"
        ).decode("utf-8")

    # Either save the result to a file or return it as a string.
    if file:
        if isinstance(file, str):  # Path
            if pathlib.Path(file).is_dir():
                raise ValueError(f"{file} is a directory.")
            else:
                with open(file, "w+") as file_handle:
                    file_handle.write(result)
        else:  # File-like object
            if "write" not in file.__dir__():
                raise TypeError(
                    f"{file} is neither a path" f"or a file-like object."
                )
            else:
                file.write(result)
    else:
        return result



"""Module aimed at providing an ontology lager on top of the RDF layer."""

from simphony_osp.session.session import Session

__all__ = ["Session", "core_session"]

core_session = Session.get_default_session()



"""Abstract superclass of any entity in the ontology."""
from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from rdflib import Graph, Literal, URIRef
from rdflib.term import Identifier

from simphony_osp.utils.cache import lru_cache_timestamp
from simphony_osp.utils.datatypes import UID, Triple

if TYPE_CHECKING:
    from simphony_osp.ontology.namespace import OntologyNamespace
    from simphony_osp.ontology.operations.container import Container
    from simphony_osp.session.session import Session
    from simphony_osp.session.wrapper import Wrapper

logger = logging.getLogger(__name__)


class OntologyEntity(ABC):
    """Abstract superclass of any entity in the ontology."""

    rdf_type: Optional[Union[URIRef, Set[URIRef]]] = None
    rdf_identifier: Type

    # Public API
    # ↓ ------ ↓

    @property
    def iri(self) -> URIRef:
        """IRI of the Entity.

        Raises:
            TypeError: When the identifier of the ontology entity is not an
                IRI.
        """
        return self.uid.to_iri()

    @property
    def identifier(self) -> Identifier:
        """Semantic web resource identifying the entity.

        Usually an URIRef or BNode.
        """
        return self.uid.to_identifier()

    @property
    def uid(self) -> UID:
        """Get a SimPhoNy identifier for this entity.

        The SimPhoNy identifier is known as UID. An UID is a Python class
        defined in SimPhoNy and can always be converted to a semantic web
        identifier.
        """
        return self._uid

    @property
    def label(self) -> Optional[str]:
        """Get the preferred label of this entity, if it exists.

        See the docstring for `label_literal` for more information on the
        definition of preferred label.
        """
        label_literal = self.label_literal
        return str(label_literal) if label_literal is not None else None

    @label.setter
    def label(self, value: str) -> None:
        """Replace the preferred label of this entity.

        When such preferred label does not exist, it is created.

        See the docstring for `label_literal` for more information on the
        definition of preferred label.
        """
        label_literal = self.label_literal
        language = (
            label_literal.language if label_literal is not None else None
        )
        self.label_literal = (
            Literal(value, lang=language) if value is not None else None
        )

    @property
    def label_lang(self) -> Optional[str]:
        """Get the language of the preferred label of this entity.

        See the docstring for `label_literal` for more information on the
        definition of preferred label.
        """
        label_literal = self.label_literal
        return label_literal.language if label_literal is not None else None

    @label_lang.setter
    def label_lang(self, value: str) -> None:
        """Set the language of the preferred label of this entity.

        See the docstring for `label_literal` for more information on the
        definition of preferred label.
        """
        self.label_literal = Literal(self.label_literal, lang=value)

    @property
    @lru_cache_timestamp(lambda self: self.session.entity_cache_timestamp)
    def namespace(self) -> Optional[OntologyNamespace]:
        """Return the ontology namespace to which this entity is associated."""
        return next((x for x in self.session.namespaces if self in x), None)

    @property
    def session(self) -> Session:
        """The session where the entity is stored."""
        return self._session

    @session.setter
    def session(self, value: Session) -> None:
        """Change the session where the entity is stored.

        Equivalent to removing the item from the previous session and adding it
        to the new session.
        """
        value.update(self)
        if self._session is not value:
            self._session.delete(self)
        self._session = value

    @property
    @lru_cache_timestamp(lambda self: self.session.entity_cache_timestamp)
    def direct_superclasses(
        self: ONTOLOGY_ENTITY,
    ) -> FrozenSet[ONTOLOGY_ENTITY]:
        """Get the direct superclasses of the entity.

        Returns:
            The direct superclasses of the entity.
        """
        return frozenset(self._get_direct_superclasses())

    @property
    @lru_cache_timestamp(lambda self: self.session.entity_cache_timestamp)
    def direct_subclasses(self: ONTOLOGY_ENTITY) -> FrozenSet[ONTOLOGY_ENTITY]:
        """Get the direct subclasses of the entity.

        Returns:
            The direct subclasses of the entity.
        """
        return frozenset(self._get_direct_subclasses())

    @property
    @lru_cache_timestamp(lambda self: self.session.entity_cache_timestamp)
    def superclasses(self: ONTOLOGY_ENTITY) -> FrozenSet[ONTOLOGY_ENTITY]:
        """Get the superclass of the entity.

        Returns:
            The direct superclasses of the entity.

        """
        return frozenset(self._get_superclasses())

    @property
    @lru_cache_timestamp(lambda self: self.session.entity_cache_timestamp)
    def subclasses(self: ONTOLOGY_ENTITY) -> FrozenSet[ONTOLOGY_ENTITY]:
        """Get the subclasses of the entity.

        Returns:
            The direct subclasses of the entity

        """
        return frozenset(self._get_subclasses())

    def is_superclass_of(self, other: OntologyEntity) -> bool:
        """Perform a superclass check.

        Args:
            other: The other ontology entity.

        Returns:
            Whether self is a superclass of the other other entity.
        """
        return self in other.superclasses

    def is_subclass_of(self, other: OntologyEntity) -> bool:
        """Perform a subclass check.

        Args:
            other: The other entity.

        Returns:
            bool: Whether self is a subclass of the other entity.

        """
        return self in other.subclasses

    def __str__(self) -> str:
        """Transform the entity into a human-readable string."""
        return (
            f"{self.label}"
            if hasattr(self, "label") and self.label is not None
            else f"{self._uid}"
        )

    def __repr__(self) -> str:
        """Transform the entity into a string."""
        header = f"{self.__class__.__name__}"
        elements = [
            f"{self.label}"
            if hasattr(self, "label") and self.label is not None
            else None,
            f"{self.uid}",
        ]
        elements = filter(lambda x: x is not None, elements)
        return f"<{header}: {' '.join(elements)}>"

    def __eq__(self, other: OntologyEntity) -> bool:
        """Check whether two entities are the same.

        Args:
            other: The other entity.

        Returns:
            bool: Whether the two entities are the same.
        """
        # TODO: Blank nodes with different IDs.
        return (
            isinstance(other, OntologyEntity)
            and self.session == other.session
            and self.identifier == other.identifier
        )

    def __hash__(self) -> int:
        """Make the entity hashable."""
        return hash((self._uid, self.session))

    def __bool__(self):
        """Returns the boolean value of the entity, always true."""
        return True

    # ↑ ------ ↑
    # Public API

    @property
    def label_literal(self) -> Optional[Literal]:
        """Get the preferred label for this entity.

        The labels are first sorted by the property defining them (which is
        an attribute of the session that this entity is stored on), and then by
        their length.

        Returns:
            The first label in the resulting ordering is returned. If the
            entity has no label, then None is returned.
        """
        labels = self.iter_labels(return_literal=True, return_prop=True)
        labels = self._sort_labels_and_properties_by_preference(labels)
        # Return the first label
        return labels[0][0] if len(labels) > 0 else None

    @label_literal.setter
    def label_literal(self, value: Optional[Literal]) -> None:
        """Replace the preferred label for this entity.

        The labels are first sorted by the property defining them (which is
        an attribute of the session that this entity is stored on), and then by
        their length.

        Args:
            value: the preferred label to replace the current one with. If
                None, then all labels for this entity are deleted.
        """
        labels = self.iter_labels(return_literal=True, return_prop=True)
        labels = self._sort_labels_and_properties_by_preference(labels)

        preferred_label = labels[0] if len(labels) > 0 else None

        # Label deletion.
        if value is None:
            for label_prop in self.session.label_properties:
                self.session.graph.remove((self.identifier, label_prop, None))
        elif preferred_label is not None:
            self.session.graph.remove(
                (self.identifier, preferred_label[1], preferred_label[0])
            )

        # Label creation.
        if value is not None:
            if preferred_label is not None:
                self.session.graph.add(
                    (self.identifier, preferred_label[1], value)
                )
            else:
                self.session.graph.add(
                    (self.identifier, self.session.label_properties[0], value)
                )

    def _sort_labels_and_properties_by_preference(
        self, labels: Iterator[Tuple[Literal, URIRef]]
    ) -> List[Tuple[Literal, URIRef]]:
        """Sort the labels for this entity in order of preference.

        The labels are first sorted by the property defining them (which is
        an attribute of the session that this entity is stored on),
        then by their language, and then by their length.

        Args:
            labels: an iterator of tuples where the first element is an
                assigned label literal (the label) and the second one the
                property used for this assignment.
        """
        # Sort by label property preference, and length.
        labels = sorted(
            labels,
            key=lambda x: (
                self.session.label_properties.index(x[1]),
                (
                    self.session.label_languages + ("en", None, x[0].language)
                ).index(x[0].language),
                len(x[0]),
            ),
        )
        return labels

    def iter_labels(
        self,
        lang: Optional[str] = None,
        return_prop: bool = False,
        return_literal: bool = True,
    ) -> Iterator[
        Union[Literal, str, Tuple[str, URIRef], Tuple[Literal, URIRef]]
    ]:
        """Returns all the available labels for this ontology entity.

        Args:
            lang: retrieve labels only in a specific language.
            return_prop: Whether to return the property that designates the
                label. When active, it is the second argument.
            return_literal: Whether to return a literal or a string with the
                label (the former contains the language, the latter not).

        Returns:
            An iterator yielding strings or literals; or tuples whose first
            element is a string or literal, and second element the property
            defining this label.
        """
        return self.session.iter_labels(
            entity=self,
            lang=lang,
            return_literal=return_literal,
            return_prop=return_prop,
        )

    @property
    def triples(self) -> Set[Triple]:
        """Get the all the triples where the entity is the subject."""
        if self.__graph is not None:
            return set(self.__graph.triples((None, None, None)))
        else:
            return set(
                self.session.graph.triples((self.identifier, None, None))
            )

    @property
    def graph(self) -> Graph:
        """Graph where the ontology entity's data lives."""
        return self.session.graph if self.session is not None else self.__graph

    @abstractmethod
    def _get_direct_superclasses(
        self: ONTOLOGY_ENTITY,
    ) -> Iterable[ONTOLOGY_ENTITY]:
        """Direct superclass getter specific to the type of ontology entity."""
        pass

    @abstractmethod
    def _get_direct_subclasses(
        self: ONTOLOGY_ENTITY,
    ) -> Iterable[ONTOLOGY_ENTITY]:
        """Direct subclass getter specific to the type of ontology entity."""
        pass

    @abstractmethod
    def _get_superclasses(self: ONTOLOGY_ENTITY) -> Iterable[ONTOLOGY_ENTITY]:
        """Superclass getter specific to the type of ontology entity."""
        pass

    @abstractmethod
    def _get_subclasses(self: ONTOLOGY_ENTITY) -> Iterable[ONTOLOGY_ENTITY]:
        """Subclass getter specific to the type of ontology entity."""
        pass

    __graph: Optional[Graph] = None  # Only exists during initialization.

    @abstractmethod
    def __init__(
        self,
        uid: UID,
        session: Optional[Union[Session, Container, Wrapper]] = None,
        triples: Optional[Iterable[Triple]] = None,
        merge: Optional[bool] = False,
    ) -> None:
        """Initialize the ontology entity.

        Args:
            uid: UID identifying the entity.
            session: Session where the entity is stored.
            triples: Construct the entity with the provided triples.
            merge: Whether overwrite the potentially existing entity in the
                session with the provided triples or just merge them with
                the existing ones.
        """
        if uid is None:
            uid = UID()
        elif not isinstance(uid, UID):
            raise Exception(
                f"Tried to initialize an ontology entity with "
                f"uid {uid}, which is not a UID object."
            )
        self._uid = uid

        # While the entity is being initialized, it belongs to no session.
        # The extra triples are added to the `__graph` attribute. While such
        # attribute exists, it is the preferred way to access the entity's
        # triples using the `triples` property.
        self._session = None
        if triples is not None:
            self.__graph = Graph()
            for s, p, o in triples:
                if s != self.identifier:
                    raise ValueError(
                        "Trying to add extra triples to an "
                        "ontology entity with a subject that "
                        "does not match the individual's "
                        "identifier."
                    )
                self.__graph.add((s, p, o))

        from simphony_osp.session.wrapper import Wrapper

        if session is None:
            from simphony_osp.ontology.operations.container import (
                ContainerEnvironment,
            )
            from simphony_osp.session.session import Environment, Session

            environment = Environment.get_default_environment()
            session = Session.get_default_session()
            if isinstance(environment, ContainerEnvironment):
                environment.connect(self.identifier)
        elif isinstance(session, Wrapper):
            session = session.session
        if self.__graph is not None:
            # Only change what is stored in the session if custom triples were
            # provided.
            if merge is False:
                session.update(self)
            elif merge is True:
                session.merge(self)
            # Otherwise, it is None -> do not change what is stored.
        self._session = session
        self.__graph = None


ONTOLOGY_ENTITY = TypeVar("ONTOLOGY_ENTITY", bound=OntologyEntity)



"""Ontology module for the SimPhoNy OSP."""

from simphony_osp.ontology.annotation import OntologyAnnotation
from simphony_osp.ontology.attribute import OntologyAttribute
from simphony_osp.ontology.composition import Composition
from simphony_osp.ontology.entity import OntologyEntity
from simphony_osp.ontology.individual import OntologyIndividual
from simphony_osp.ontology.namespace import OntologyNamespace
from simphony_osp.ontology.oclass import OntologyClass
from simphony_osp.ontology.relationship import OntologyRelationship
from simphony_osp.ontology.restriction import Restriction

__all__ = [
    "Composition",
    "Restriction",
    "OntologyAnnotation",
    "OntologyAttribute",
    "OntologyEntity",
    "OntologyIndividual",
    "OntologyNamespace",
    "OntologyClass",
    "OntologyRelationship",
]



"""Restrictions on ontology classes."""
from __future__ import annotations

import logging
from enum import Enum
from typing import TYPE_CHECKING, Iterable, Optional, Tuple, Union

from rdflib import OWL, BNode, URIRef
from rdflib.term import Identifier

from simphony_osp.ontology.attribute import OntologyAttribute
from simphony_osp.ontology.entity import OntologyEntity
from simphony_osp.ontology.relationship import OntologyRelationship
from simphony_osp.utils.datatypes import UID, Triple

if TYPE_CHECKING:
    from simphony_osp.ontology.oclass import OntologyClass
    from simphony_osp.session.session import Session

logger = logging.getLogger(__name__)


class QUANTIFIER(Enum):
    """The different quantifiers for restrictions."""

    SOME = 1
    ONLY = 2
    EXACTLY = 3
    MIN = 4
    MAX = 5
    VALUE = 6


class RTYPE(Enum):
    """The two types of restrictions."""

    ATTRIBUTE_RESTRICTION = 1
    RELATIONSHIP_RESTRICTION = 2


class Restriction(OntologyEntity):
    """A class to represent restrictions on ontology classes."""

    rdf_type = OWL.Restriction
    rdf_identifier = BNode

    def __init__(
        self,
        uid: UID,
        session: Optional[Session] = None,
        triples: Optional[Iterable[Triple]] = None,
        merge: bool = False,
    ) -> None:
        """Initialize the restriction class.

        Args:
            uid: An UID whose data attribute is the blank node that represents
                the restriction.
            session: Session where the restriction is stored.
            triples: Construct the restriction with the provided triples.
            merge: Whether overwrite the potentially existing entity in the
                session with the provided triples or just merge them with
                the existing ones.
        """
        if not isinstance(uid.data, BNode):
            raise ValueError(
                f"Restrictions are anonymous class descriptions, "
                f"and thus, they can only have blank nodes as "
                f"UID, not {type(uid.data)}."
            )
        super().__init__(uid, session, triples, merge=merge)

    # Public API
    # ↓ ------ ↓

    def __str__(self) -> str:
        """Transform to string."""
        return " ".join(
            map(str, (self._property, self.quantifier, self.target))
        )

    @property
    def quantifier(self) -> QUANTIFIER:
        """Get the quantifier of the restriction.

        Returns:
            QUANTIFIER: The quantifier of the restriction.
        """
        quantifier, _ = self._get_quantifier_and_target()
        return quantifier

    @property
    def target(self) -> Union[OntologyClass, URIRef]:
        """The target ontology class or datatype.

        Returns:
            The target class or datatype.
        """
        _, target = self._get_quantifier_and_target()
        try:
            target = self.session.from_identifier(target)
        except KeyError:
            pass
        return target

    @property
    def relationship(self) -> OntologyRelationship:
        """The relationship the RELATIONSHIP_RESTRICTION acts on.

        Raises:
            AttributeError: Called on an ATTRIBUTE_RESTRICTION.

        Returns:
            The relationship the restriction acts on.
        """
        if self.rtype == RTYPE.ATTRIBUTE_RESTRICTION:
            raise AttributeError
        return self._property

    @property
    def attribute(self) -> OntologyAttribute:
        """The attribute the restriction acts on.

        Only for ATTRIBUTE_RESTRICTIONs.

        Raises:
            AttributeError: self is a RELATIONSHIP_RESTRICTIONs.

        Returns:
            The attribute.
        """
        if self.rtype == RTYPE.RELATIONSHIP_RESTRICTION:
            raise AttributeError
        return self._property

    @property
    def rtype(self) -> RTYPE:
        """Return the type of restriction.

        Whether the restriction acts on attributes or relationships.

        Returns:
            RTYPE: The type of restriction.
        """
        prop = self._property
        if isinstance(prop, OntologyRelationship):
            return RTYPE.RELATIONSHIP_RESTRICTION
        elif isinstance(prop, OntologyAttribute):
            return RTYPE.ATTRIBUTE_RESTRICTION
        else:
            raise RuntimeError(
                f"Invalid property {prop} for restriction. "
                f"{OntologyRelationship} or {OntologyAttribute}"
                f" were expected."
            )

    # ↑ ------ ↑
    # Public API

    def _get_quantifier_and_target(
        self,
    ) -> Tuple[Optional[QUANTIFIER], Optional[Identifier]]:
        """Get both the quantifier and the target of the restriction.

        Since the calculation of one involves the calculation of the other,
        this function returns both, and caches the result. Then,
        then each individual element can be queried through the properties
        `quantifier` and `target`.

        Returns:
            A tuple where the first element is a quantifier (if any) and
            the second element the identifier of the target (if any).
        """
        for predicate, quantifier in [
            (OWL.someValuesFrom, QUANTIFIER.SOME),
            (OWL.allValuesFrom, QUANTIFIER.ONLY),
            (OWL.cardinality, QUANTIFIER.EXACTLY),
            (OWL.minCardinality, QUANTIFIER.MIN),
            (OWL.maxCardinality, QUANTIFIER.MAX),
            (OWL.hasValue, QUANTIFIER.VALUE),
        ]:
            x = self.session.graph.value(self.identifier, predicate)
            if x:
                return quantifier, x
        else:
            return None, None

    @property
    def _property(self) -> Union[OntologyRelationship, OntologyAttribute]:
        """The relationship or attribute the restriction acts on.

        Returns:
            Union[OntologyRelationship, OntologyAttribute]:
                object of owl:onProperty predicate.
        """
        prop = self.session.graph.value(self.identifier, OWL.onProperty)
        if prop and not isinstance(prop, BNode):
            # TODO: handle inverse properties defined as blank nodes.
            return self.session.from_identifier(prop)
        else:
            raise RuntimeError(
                f"Property {prop} is not within any installed " f"ontology."
            )

    def _get_direct_superclasses(self) -> Iterable[OntologyEntity]:
        """Restrictions have no superclasses."""
        return iter(())

    def _get_direct_subclasses(self) -> Iterable[OntologyEntity]:
        """Restrictions have no subclasses."""
        return iter(())

    def _get_superclasses(self) -> Iterable[OntologyEntity]:
        """Restrictions have no superclasses."""
        return iter(())

    def _get_subclasses(self) -> Iterable[OntologyEntity]:
        """Restrictions have no subclasses."""
        return iter(())



"""This files defines composition of classes."""
from __future__ import annotations

import logging
from enum import Enum
from typing import TYPE_CHECKING, Iterable, List, Optional, Tuple, Union

from rdflib import OWL, RDF, BNode

from simphony_osp.ontology.entity import OntologyEntity
from simphony_osp.utils.datatypes import UID, Triple

if TYPE_CHECKING:
    from simphony_osp.ontology.oclass import OntologyClass
    from simphony_osp.ontology.restriction import Restriction
    from simphony_osp.session.session import Session

logger = logging.getLogger(__name__)


class OPERATOR(Enum):
    """Operations applicable to class definitions."""

    AND = 1  # owl:intersectionOf
    OR = 2  # owl:unionOf
    NOT = 3  # owl:complementOf


class Composition(OntologyEntity):
    """Combine multiple classes using logical formulae."""

    rdf_type = OWL.Class
    rdf_identifier = BNode

    # Public API
    # ↓ ------ ↓

    @property
    def operator(self) -> OPERATOR:
        """The operator that connects the different classes in the formula.

        Returns:
            The operator Enum.
        """
        operator, _ = self._get_operator_and_operands
        return operator

    @property
    def operands(
        self,
    ) -> Tuple[Union[OntologyClass, Composition, Restriction]]:
        """The individual classes the formula is composed of.

        Returns:
            The operands.
        """
        _, operands = self._get_operator_and_operands
        return tuple(operands)

    def __str__(self) -> str:
        """Transform to a Protege-like string."""
        s = f" {self.operator} ".join(map(str, self.operands))
        if self.operator == OPERATOR.NOT:
            s = f"{self.operator} {s}"
        return f"({s})"

    # ↑ ------ ↑
    # Public API

    def __init__(
        self,
        uid: UID,
        session: Optional[Session] = None,
        triples: Optional[Iterable[Triple]] = None,
        merge: bool = False,
    ) -> None:
        """Initialize the class composition."""
        if not isinstance(uid.data, BNode):
            raise ValueError(
                f"Compositions are anonymous class descriptions, "
                f"and thus, they can only have blank nodes as "
                f"UIDs, not {type(uid.data)}."
            )
        super().__init__(uid, session, triples, merge=merge)

    def _get_operator_and_operands(
        self,
    ) -> Tuple[
        Optional[OPERATOR],
        List[Union[OntologyClass, Composition, Restriction]],
    ]:
        """Look up operator and operands in the graph."""
        for predicate, operator in [
            (OWL.unionOf, OPERATOR.OR),
            (OWL.intersectionOf, OPERATOR.AND),
            (OWL.complementOf, OPERATOR.NOT),
        ]:
            operands = []
            o = self.session.graph.value(self.identifier, predicate)
            if operator == OPERATOR.NOT:
                operand = self._get_operand(o)
                operands += [operand] if operand is not None else []
                return operator, operands
            x = self.session.graph.value(o, RDF.first)
            while x:
                operand = self._get_operand(x)
                operands += [operand] if operand is not None else []
                o = self.session.graph.value(o, RDF.rest)
                x = self.session.graph.value(o, RDF.first)
            return operator, operands

    def _get_operand(
        self, identifier
    ) -> Optional[Union[OntologyClass, Composition, Restriction]]:
        """Get an operand to the from an identifier."""
        try:
            operand = self.session.from_identifier(identifier)
            operand: Union[OntologyClass, Composition, Restriction]
            return operand
        except KeyError:
            pass

    def _get_direct_superclasses(self) -> Iterable[Composition]:
        """Compositions have no superclasses."""
        return iter(())

    def _get_direct_subclasses(self) -> Iterable[Composition]:
        """Compositions have no subclasses."""
        return iter(())

    def _get_superclasses(self) -> Iterable[Composition]:
        """Compositions have no superclasses."""
        return iter(())

    def _get_subclasses(self) -> Iterable[Composition]:
        """Compositions have no subclasses."""
        return iter(())



"""An ontology namespace."""
from __future__ import annotations

import itertools
import logging
from itertools import chain
from typing import TYPE_CHECKING, Any, Iterable, Iterator, Optional, Set, Union

from rdflib import URIRef
from rdflib.term import Identifier

from simphony_osp.ontology.entity import OntologyEntity
from simphony_osp.utils.cache import lru_cache_timestamp

if TYPE_CHECKING:
    from simphony_osp.session.session import Session


logger = logging.getLogger(__name__)


class OntologyNamespace:
    """An ontology namespace."""

    # Public API
    # ↓ ------ ↓

    def __init__(
        self,
        iri: Union[str, URIRef],
        ontology: Optional[Session] = None,
        name: Optional[str] = None,
    ):
        """Initialize a namespace object.

        Args:
            iri: The IRI of the namespace.
            ontology: The ontology to which the namespace is connected.
            name: The name of the namespace
        """
        from simphony_osp.session.session import Session

        ontology = ontology or Session.get_default_session()
        self._iri = URIRef(iri)
        self._ontology = ontology
        ontology.bind(name, iri)

    @property
    def name(self) -> Optional[str]:
        """The name of the namespace."""
        return self.ontology.get_namespace_bind(self)

    @property
    def iri(self) -> URIRef:
        """The IRI of the namespace."""
        return self._iri

    @property
    def ontology(self) -> Session:
        """Returns the session that the namespace is bound to."""
        return self._ontology

    def __eq__(self, other: OntologyNamespace) -> bool:
        """Check whether the two namespace objects are equal.

        Args:
            other: The namespace object to compare with.

        Returns:
            bool: Whether the given namespace object is equal.
        """
        return (
            isinstance(other, OntologyNamespace)
            and self.ontology is other.ontology
            and self.iri == other.iri
        )

    def __hash__(self) -> int:
        """Hash the namespace.

        The namespace is defined by its IRI and its underlying data
        structure (the ontology), which are immutable attributes.
        """
        return hash((self.ontology, self.iri))

    def __getattr__(self, name: str) -> OntologyEntity:
        """Get an entity from the ontology associated to the namespace.

        Args:
            name: The label or namespace suffix of the ontology entity.

        Raises:
            AttributeError: Unknown label or suffix.
            AttributeError: Multiple entities for the given label or suffix.

        Returns:
            An ontology entity with matching label or suffix.
        """
        try:
            return self.get(name)
        except KeyError as e:
            raise AttributeError(str(e)) from e

    def __getitem__(self, label: str) -> OntologyEntity:
        """Get an entity from the ontology associated to the namespace.

        Useful for entities whose labels or suffixes contain characters which
        are not compatible with the Python syntax.

        Args:
            label: The label of the ontology entity.

        Raises:
            KeyError: Unknown label or suffix.
            KeyError: Multiple entities for the given label or suffix.

        Returns:
            An ontology entity with matching label or suffix.
        """
        if not isinstance(label, str):
            exception = TypeError(
                f"{str(type(self)).capitalize()} indices must be"
                f"of type {str}."
            )
            raise exception

        return self.get(label)

    def __dir__(self) -> Iterable[str]:
        """Attributes available for the OntologyNamespace class.

        Returns:
            The available attributes, which include the methods and
            the ontology entities in the namespace, both by label and suffix.
        """
        entity_autocompletion = chain(
            self._iter_labels(), self._iter_suffixes()
        )
        entity_autocompletion = filter(
            lambda x: x.isidentifier(), entity_autocompletion
        )
        return chain(super().__dir__(), entity_autocompletion)

    def _ipython_key_completions_(self):
        """Items available for the OntologyNamespace class.

        Returns:
            The available ontology entities in the namespace, both by label
            and suffix.
        """
        entity_autocompletion = chain(
            self._iter_labels(), self._iter_suffixes()
        )
        super_completion = getattr(super(), "_ipython_key_completions_", set())
        return itertools.chain(entity_autocompletion, super_completion)

    def __iter__(self) -> Iterator[OntologyEntity]:
        """Iterate over the ontology entities in the namespace."""
        return (entity for entity in iter(self.ontology) if entity in self)

    def __contains__(self, item: Union[OntologyEntity, Identifier]) -> bool:
        """Check whether the given entity is part of the namespace.

        Args:
            item: An ontology entity or identifier.

        Returns:
            Whether the given entity name or IRI is part of the namespace.
            Blank nodes are never part of a namespace.
        """
        if isinstance(item, Identifier) and not isinstance(item, URIRef):
            return False
        elif isinstance(item, URIRef):
            return item.startswith(self.iri)
        elif isinstance(item, OntologyEntity):
            return (
                item.identifier in self
                if item.session is self.ontology
                else False
            )
        else:
            raise TypeError(
                f"in {type(self)} requires, "
                f"{Identifier} or {OntologyEntity} as left "
                f"operand, not {type(item)}."
            )

    def __len__(self) -> int:
        """Return the number of entities in the namespace."""
        return sum(1 for _ in self)

    @lru_cache_timestamp(
        lambda self: self.ontology.entity_cache_timestamp, maxsize=4096
    )
    def get(self, name: str, default: Optional[Any] = None) -> OntologyEntity:
        """Get ontology entities from the registry by suffix or label.

        Args:
            name: The label or suffix of the ontology entity.
            default: The value to return if no entity is found.

        Raises:
            KeyError: Unknown label or suffix.
            KeyError: Multiple entities for the given label or suffix.

        Returns:
            The ontology entity with given label or suffix, or the default
            value.
        """
        from_label = self._from_label_set(name, case_sensitive=True)

        try:
            from_suffix = self.from_suffix(name)
        except (KeyError, ValueError):
            from_suffix = None

        entities = from_label | ({from_suffix} if from_suffix else set())
        num_entities = len(entities)
        if num_entities == 0:
            if default is None:
                raise KeyError(
                    f"No entity with label or suffix {name} was found in "
                    f"namespace {self}."
                )
            else:
                entities = {default}
        elif num_entities >= 2:
            error = (
                f"There are multiple entities with label or suffix {name} "
                f"in namespace {self}: {', '.join(r.iri for r in entities)}. "
                f"Please refer to a specific element of the "
                f"list by calling the method `from_iri(iri)` from "
                f"namespace {self}."
            )
            raise KeyError(error)
        return entities.pop()

    def from_suffix(self, suffix: str) -> OntologyEntity:
        """Get an ontology entity from its namespace suffix.

        Args:
            suffix: Suffix of the ontology entity.

        Raises:
            KeyError: When no entity with such suffix exists in the namespace.
            ValueError: When an invalid suffix is received (e.g. it contains a
                space).
        """
        return self.from_iri(str(self._iri) + suffix)

    @lru_cache_timestamp(
        lambda self: self.ontology.entity_cache_timestamp, maxsize=4096
    )
    def from_iri(self, iri: Union[str, URIRef]) -> OntologyEntity:
        """Get an ontology entity directly from its IRI.

        For consistency, this method only returns entities from this namespace.

        Args:
            iri: The iri of the ontology entity.

        Returns:
            The ontology entity.

        Raises:
            KeyError: When the IRI does not belong to the namespace.
            ValueError: When an invalid IRI is received.
        """
        if " " in iri:
            raise ValueError(f"Invalid IRI: {iri}.")
        # TODO: check that it is a valid URI instead of just checking if it
        #  has spaces. The check is needed to avoid RDFLib complaining that
        #  trying to serialize an IRI with spaces will break.

        iri = URIRef(str(iri))
        if iri in self:
            return self.ontology.from_identifier(iri)
        else:
            raise KeyError(
                f"The IRI {iri} does not belong to the namespace" f"{self}."
            )

    @lru_cache_timestamp(
        lambda self: self.ontology.entity_cache_timestamp, maxsize=4096
    )
    def from_label(
        self,
        label: str,
        lang: Optional[str] = None,
        case_sensitive: bool = False,
    ) -> OntologyEntity:
        """Get an ontology entity from its label.

        Args:
            label: The label to match.
            lang: Optionally filter labels by a specific language.
            case_sensitive: Whether the match should be case-sensitive or
                not. The default setting is a case-insensitive lookup.

        Raises:
            KeyError: No label matches the given one.
            KeyError: More than one label matches the given one.
        """
        try:
            entities = {
                entity
                for entity in self.ontology.from_label(
                    label, lang=lang, case_sensitive=case_sensitive
                )
                if entity in self
            }
        except KeyError:
            entities = set()
        if len(entities) == 0:
            error = "No element with label %s was found in namespace %s." % (
                label,
                self,
            )
            raise KeyError(error)
        elif len(entities) >= 2:
            element_suffixes = (r.iri[len(self.iri) :] for r in entities)
            error = (
                f"There are multiple elements "
                f"({', '.join(element_suffixes)}) with label"
                f" {label} for namespace {self}."
                f"\n"
                f"Please refer to a specific element of the "
                f"list by calling from_iri(IRI) for "
                f"namespace {self} for one of the following "
                f"IRIs: " + "{iris}."
            ).format(iris=", ".join(entity.iri for entity in entities))
            raise KeyError(error)
        return entities.pop()

    # ↑ ------ ↑
    # Public API

    def __str__(self) -> str:
        """Transform the namespace to a human-readable string.

        Returns:
            The resulting string.
        """
        return f"{self.name} ({self.iri})"

    def __repr__(self) -> str:
        """Transform the namespace to a string.

        Returns:
            The resulting string.
        """
        return f"<{self.name}: {self.iri}>"

    @lru_cache_timestamp(
        lambda self: self.ontology.entity_cache_timestamp, maxsize=4096
    )
    def _from_label_set(
        self,
        label: str,
        lang: Optional[str] = None,
        case_sensitive: bool = False,
    ) -> Set[OntologyEntity]:
        """Get ontology entities from their labels.

        Args:
            label: The label to match.
            lang: Optionally filter labels by a specific language.
            case_sensitive: Whether the match should be case-sensitive or
                not. The default setting is a case-insensitive lookup.
        """
        try:
            entities = {
                entity
                for entity in self.ontology.from_label(
                    label, lang=lang, case_sensitive=case_sensitive
                )
                if entity in self
            }
        except KeyError:
            entities = set()

        return entities

    def _iter_labels(self) -> Iterator[str]:
        """Iterate over the labels of the ontology entities in the namespace.

        Returns:
            An iterator of strings containing the labels.
        """
        return (
            label
            for iri in self._iter_identifiers()
            for label in self.ontology.iter_labels(iri, return_literal=False)
        )

    def _iter_suffixes(self) -> Iterator[str]:
        """Iterate over suffixes of the ontology entities in the namespace.

        Returns:
            An iterator of suffixes containing the suffixes.
        """
        return (str(iri)[len(str(self._iri)) :] for iri in self._iter_iris())

    def _iter_identifiers(self) -> Iterator[Identifier]:
        """Iterate over the identifiers of the ontology entities.

        For consistency, only returns ontology entities that belong to the
        namespace.

        Returns:
            An iterator of identifiers.
        """
        return (
            identifier
            for identifier in self.ontology.iter_identifiers()
            if identifier in self
        )

    def _iter_iris(self) -> Iterator[URIRef]:
        """Iterate over the IRIs of the ontology entities in the namespace.

        Returns:
            An iterator of IRIs.
        """
        return (x for x in self._iter_identifiers() if isinstance(x, URIRef))



"""Interface between the SimPhoNy OSP and SQLAlchemy."""

from typing import Dict, Optional

from rdflib import Graph, URIRef
from rdflib.term import Identifier

from simphony_osp.interfaces.interface import BufferType, Interface


class SQLAlchemy(Interface):
    """An interface to an SQL database using SQLAlchemy."""

    _identifier: Identifier = URIRef(
        "https://www.simphony-project.eu/SQLAlchemy"
    )
    """The identifier of the graph serving as context of the saved triples."""

    _uri: Optional[str] = None
    """SQLAlchemy URI used to connect to the database."""

    _buffers: Optional[Dict[BufferType, Graph]]
    """Triple buffers (see docstring of `BufferType`)."""

    base: Optional[Graph] = None
    """Representation of the contents of the database as an RDFLib graph
    using the `rdflib-sqlalchemy` plug-in."""

    # Interface
    # ↓ ----- ↓

    entity_tracking: bool = False

    def open(self, configuration: str, create: bool = False):
        """Open a connection to the database.

        Args:
            configuration: The SQLAlchemy URI pointing to the database to
                which the user wishes to connect.
            create: Whether to create the database file if it does not exist.
        """
        # TODO: Create databases if create is `True`.
        if self._uri is not None and self._uri != configuration:
            raise RuntimeError(
                f"A different database {self._uri}" f"is already open!"
            )

        self.base = Graph("SQLAlchemy", identifier=self._identifier)
        self.base.open(configuration, create=True)
        self._uri = configuration

    def close(self) -> None:
        """Close the connection to the database."""
        if self.base is not None:
            self.base.close(commit_pending_transaction=False)
            self._uri = None
            self.base = None

    def commit(self):
        """Commit pending changes to the triple store."""
        # The `InterfaceDriver` will simply add the triples to the base graph
        # and commit them. Nothing to do here.
        pass

    def populate(self):
        """The base graph does not need to be populated. Nothing to do."""
        pass

    # ↑ ----- ↑



"""A user-facing class creating a session using a specific interface."""

from abc import ABC, abstractmethod
from typing import Iterable, Optional, Type, Union

from rdflib import Graph

from simphony_osp.interfaces.interface import Interface, InterfaceDriver
from simphony_osp.ontology.entity import OntologyEntity
from simphony_osp.ontology.operations.container import Container
from simphony_osp.session.session import Session


class Wrapper:
    """A user-facing class for managing a session.

    The idea is to create hybrid subclasses from this class, for example
    ontology entities that are also able to manage sessions.

    Anything that has a `_session` attribute may be used to manage a session.
    """

    # Any class combined with the wrapper is expected to provide the two
    # properties below.
    session: Session
    interface: Optional[Interface] = None
    _session: Session
    _exit_container: bool = False

    def __enter__(self):
        """Enter the associated session's context."""
        self._session.__enter__()
        if isinstance(self, Container):
            Container.__enter__(self)
        return self

    def __exit__(self, *args):
        """Exit the associated session's context."""
        if isinstance(self, Container):
            self._exit_container = True
            Container.__exit__(self, *args)
            self._exit_container = False
        self._session.__exit__(*args)

    def commit(self) -> None:
        """Commit the changes made to the backend."""
        return self._session.commit()

    def close(self) -> None:
        """Close the connection to the backend."""
        if isinstance(self, Container):
            Container.close(self)
        if not self._exit_container:
            return self._session.close()

    def delete(self, *entities: OntologyEntity) -> None:
        """Delete entities from the backend."""
        for entity in entities:
            self._session.delete(entity)

    def compute(self, *args, **kwargs) -> None:
        """Instructs the backend to run a simulation if supported."""
        return self._session.compute(*args, **kwargs)


class WrapperSpawner(ABC, Session):
    """A user-facing class for spawning a session."""

    @classmethod
    @abstractmethod
    def _get_interface(cls) -> Type[Interface]:
        """The type of interface that the instantiated session will use."""
        pass

    def __new__(
        cls,
        configuration_string: str = "",
        create: bool = False,
        ontology: Optional[Union[Session, bool]] = None,
        **kwargs: Union[
            str,
            int,
            float,
            bool,
            None,
            Iterable[Union[str, int, float, bool, None]],
        ]
    ) -> Session:
        """Initialize the session using the wrapper's interface type.

        Creates an interface and a store using that interface. Then
        initialize the session using such store.
        """
        interface_class = cls._get_interface()
        interface_instance = interface_class(**kwargs)
        store = InterfaceDriver(interface=interface_instance)
        graph = Graph(store=store)
        graph.open(configuration_string, create=create)
        session = Session(base=graph, driver=store, ontology=ontology)
        session._driver = store
        return session



"""Parses OWL ontologies."""

import io
import logging
import os.path
from typing import Dict, Optional, Set, Union

import requests
import yaml
from rdflib import RDF, RDFS, Graph, URIRef
from rdflib.util import guess_format

import simphony_osp.ontology.parser.owl.keywords as keywords
from simphony_osp.ontology.parser.parser import OntologyParser

logger = logging.getLogger(__name__)


class OWLParser(OntologyParser):
    """Parses OWL ontologies."""

    _file_path: str
    _graph: Optional[Graph] = None  # For lazy evaluation.
    _yaml_config: Union[dict, list]

    def __init__(self, path: str):
        """Initialize the OWL ontology parser."""
        self._yaml_config = self._load_yaml_config(path)
        self._file_path = self.parse_file_path(path)

    @property
    def identifier(self) -> str:
        """Get the identifier of the loaded ontology package.

        Returns:
            The identifier of the loaded ontology package.
        """
        return self._yaml_config[keywords.IDENTIFIER_KEY].lower()

    @property
    def namespaces(self) -> Dict[str, URIRef]:
        """Fetch the namespaces from the ontology files.

        Returns:
            A dictionary containing the defined namespace names and URIs.
        """
        namespaces = {
            label.lower(): URIRef(value)
            for label, value in self._yaml_config[
                keywords.NAMESPACES_KEY
            ].items()
        }
        # TODO: Do not do this, the ontology creator is responsible.
        for name, iri in namespaces.items():
            if not (iri.endswith("#") or iri.endswith("/")):
                namespaces[name] = iri + "#"
        return namespaces

    @property
    def requirements(self) -> Set[str]:
        """Fetch the dependencies from the ontology file."""
        requirements = self._yaml_config.get(keywords.REQUIREMENTS_KEY, list())
        if not isinstance(requirements, list):
            raise ValueError(
                f"Object of type {type(requirements)} given as list of "
                f"requirements for the ontology {self.identifier}. You need "
                f"to specify a list."
            )
        return set(requirements)

    @property
    def graph(self) -> Graph:
        """Fetch the ontology graph from the ontology file."""
        if not self._graph:
            file_format = self._yaml_config.get(keywords.FILE_FORMAT_KEY, None)
            self._graph = self._read_ontology_graph(
                self._yaml_config, self._file_path, file_format
            )
        return self._graph

    def install(self, destination: str):
        """Store the parsed files at the given destination.

        This function is meant to copy the ontology to the SimPhoNy data
        directory. So usually the destination will be `~/.simphony-osp`.

        Args:
            destination: the SimPhoNy data directory.
        """
        # TODO: This function is related to exporting ontologies.
        rdf_relative_path = f"{self.identifier}.xml"
        graph_destination = os.path.join(destination, rdf_relative_path)
        config_destination = os.path.join(
            destination, f"{self.identifier}.yml"
        )
        saved_graph, saved_config = (False,) * 2
        try:
            # Save graph.
            self.graph.serialize(graph_destination, format="xml")
            saved_graph = True
            # Save YML config.
            yaml_config = self._yaml_config
            yaml_config[keywords.RDF_FILE_KEY] = rdf_relative_path
            yaml_config[keywords.FILE_FORMAT_KEY] = "xml"
            with open(config_destination, "w") as file:
                yaml.safe_dump(yaml_config, file)
            saved_config = True
        finally:
            if not all((saved_graph, saved_config)):
                if saved_graph:
                    os.remove(graph_destination)
                if saved_config:
                    os.remove(config_destination)

    @classmethod
    def _load_yaml_config(cls, path: str) -> list:
        """Load the given YAML config file for the OWL ontology.

        Loads the YAML config file into the Parser object. Validates it first.

        Args:
            path: path of the YAML config file to load.

        Returns:
            doc: YAML doc for the specified config file, validated
                by `_validate_yaml_config`.

        Raises:
            SyntaxError: needed keywords not found in file.
            ValueError: invalid values for some keywords.
        """
        path = cls.parse_file_path(path)
        doc = cls.load_yaml(path)
        try:
            cls._validate_yaml_config(doc)
        except SyntaxError or ValueError as e:
            raise type(e)(f"Invalid configuration file {path}. {e}")
        return doc

    @staticmethod
    def _validate_yaml_config(doc: list):
        """Validate the given YAML config file for the OWL ontology.

        Args:
            doc: the path of the YAML config file.

        Raises:
            SyntaxError: needed keywords not found in file.
            ValueError: invalid values for some keywords.
        """
        if not all(KEY in doc for KEY in keywords.MANDATORY_KEYS):
            raise SyntaxError(
                "The given file is not a YAML configuration "
                "file for an OWL ontology. Make sure that it "
                "contains all of the following keys: %s."
                % ", ".join(KEY for KEY in keywords.MANDATORY_KEYS)
            )
        other_keys = {KEY for KEY in doc} - set(keywords.ALL_KEYS)
        if other_keys:
            raise SyntaxError(
                "Specified unknown keys in YAML file: %s" % other_keys
            )
        if "." in doc[keywords.IDENTIFIER_KEY]:
            raise ValueError(
                "No dots are allowed in the given package "
                "identifier: %s." % doc[keywords.IDENTIFIER_KEY]
            )

    @staticmethod
    def _read_ontology_graph(
        yaml_config_doc: dict,
        yaml_config_path: str,
        file_format: Optional[str] = None,
    ) -> Graph:
        """Get the ontology from the file specified in the configuration file.

        Args:
            yaml_config_doc: The YAML doc resulting from loading the YAML
                config file for OWL ontologies. The doc must have been
                validated with `_validate_yaml_config` before being passed to
                this function.
            yaml_config_path: The path where the YAML config file was
                read. It is used to resolve the relative path to the ontology
                file.
            file_format: The format of the file containing the ontology graph.
                When not provided, it will be guessed using `guess_format`
                from `rdflib.util`.

        Returns:
            Graph: The ontology graph.
        """
        rdf_file_location = yaml_config_doc[keywords.RDF_FILE_KEY]
        file_format = file_format or guess_format(rdf_file_location)
        if rdf_file_location.startswith(("http://", "https://")):
            logger.info(f"Downloading {rdf_file_location}.")
            content = requests.get(rdf_file_location).content.decode("utf-8")
            file_like = io.StringIO(content)
        else:
            rdf_file_location = os.path.join(
                os.path.dirname(yaml_config_path), rdf_file_location
            )
            file_like = open(rdf_file_location, "rb")
        graph = Graph()
        graph.parse(file_like, format=file_format)
        file_like.close()

        # Patch DCTERMS, which defines http://purl.org/dc/terms/Agent
        # both as a class and as an instance of
        # http://purl.org/dc/terms/AgentClass
        if (
            URIRef("http://purl.org/dc/terms/"),
            URIRef("http://purl.org/dc/terms/modified"),
            None,
        ) in graph:
            graph.remove(
                (
                    URIRef("http://purl.org/dc/terms/Agent"),
                    RDF.type,
                    URIRef("http://purl.org/dc/terms/AgentClass"),
                )
            )
            graph.add(
                (
                    URIRef("http://purl.org/dc/terms/Agent"),
                    RDFS.subClassOf,
                    URIRef("http://purl.org/dc/terms/AgentClass"),
                )
            )

        return graph



"""Special kind of ontology individual designed to organize entities."""
from __future__ import annotations

import logging
from typing import BinaryIO

from rdflib.term import URIRef

from simphony_osp.ontology.operations import Operations
from simphony_osp.utils.simphony_namespace import simphony_namespace

logger = logging.getLogger(__name__)

buf_size = 1024


class File(Operations):
    """Actions for ontology individuals representing a file."""

    iri: str = str(simphony_namespace.File)

    @property
    def _session(self):
        """Session of the ontology individual connected to the action."""
        return self._individual.session

    @property
    def _identifier(self) -> URIRef:
        """Identifier of the ontology individual connected to the action.

        Only `URIRef` objects are allowed.
        """
        identifier = self._individual.identifier
        if not isinstance(identifier, URIRef):
            raise TypeError(
                f"Invalid identifier type "
                f"{type(identifier)} for file individual."
            )
        return identifier

    def upload(self, path: str) -> None:
        """Upload a file.

        Queues a file to be uploaded to the server. When a commit is
        performed, the data is sent.
        """
        if self._session.driver is not None:
            file = open(path, "rb")
            self._session.driver.queue(self._identifier, file)
        else:
            logger.warning(
                "This session does not support saving new files. The "
                "contents of the file will NOT be saved."
            )

    def download(self, path: str) -> None:
        """Download the file."""
        if self._session.driver is not None:
            with self._session.driver.load(self._identifier) as file:
                with open(path, "wb") as new_file:
                    data = True
                    while data:
                        data = file.read(buf_size)
                        new_file.write(data)
        else:
            raise FileNotFoundError(
                "This session does not support file storage. Unable to "
                "retrieve the file contents."
            )

    @property
    def handle(self) -> BinaryIO:
        """Get a file handle to operate with."""
        if self._session.driver is not None:
            return self._session.driver.load(self._identifier)
        else:
            raise FileNotFoundError(
                "This session does not support file storage. Unable to "
                "retrieve the file contents."
            )

    def overwrite(self, contents: BinaryIO) -> None:
        """Overwrite the file contents with a byte stream."""
        if self._session.driver is not None:
            self._session.driver.queue(self._identifier, contents)
        else:
            logger.warning(
                "This session does not support saving new files. The "
                "contents of the file will NOT be saved."
            )



"""Module defining operations for ontology individuals.

Interactive ontology individuals add additional functionality to individuals of
a specific class. For example, SimPhoNY Files have a method to download or
upload an associated file object.
"""

from simphony_osp.ontology.operations.operations import (
    Operations,
    find_operations,
)

__all__ = ["Operations", "find_operations"]



"""A relationship defined in the ontology."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Iterable, Iterator, Optional

from rdflib import OWL, RDF, RDFS
from rdflib.term import Identifier

from simphony_osp.ontology.entity import OntologyEntity
from simphony_osp.utils.datatypes import UID, Triple

if TYPE_CHECKING:
    from simphony_osp.session.session import Session

logger = logging.getLogger(__name__)

# TODO characteristics


class OntologyRelationship(OntologyEntity):
    """A relationship defined in the ontology."""

    rdf_type = OWL.ObjectProperty
    rdf_identifier = Identifier

    # ↓ --------------------- Public API --------------------- ↓ #

    @property
    def inverse(self) -> Optional[OntologyRelationship]:
        """Get the inverse relationship if it exists."""
        inverse = self.session.graph.objects(self.identifier, OWL.inverseOf)
        inverse = next(iter(inverse), None)
        inverse = (
            self.session.from_identifier(inverse)
            if inverse is not None
            else None
        )
        return inverse

    # ↑ --------------------- Public API --------------------- ↑ #

    def __init__(
        self,
        uid: UID,
        session: Optional[Session] = None,
        triples: Optional[Iterable[Triple]] = None,
        merge: bool = False,
    ) -> None:
        """Initialize the ontology relationship.

        Args:
            uid: UID identifying the entity.
            session: Session where the entity is stored.
            triples: Construct the relationship with the provided triples.
            merge: Whether overwrite the potentially existing entity in the
                session with the provided triples or just merge them with
                the existing ones.
        """
        super().__init__(uid, session, triples, merge=merge)
        logger.debug("Create ontology relationship %s" % self)

    def _get_direct_superclasses(self) -> Iterator[OntologyRelationship]:
        """Get all the direct superclasses of this relationship.

        Returns:
            The direct superrelationships.
        """
        return (
            self.session.from_identifier_typed(o, typing=OntologyRelationship)
            for o in self.session.graph.objects(self.iri, RDFS.subPropertyOf)
        )

    def _get_direct_subclasses(self) -> Iterator[OntologyRelationship]:
        """Get all the direct subclasses of this relationship.

        Returns:
            OntologyRelationship: The direct subrelationships
        """
        return (
            self.session.from_identifier_typed(s, typing=OntologyRelationship)
            for s in self.session.graph.subjects(RDFS.subPropertyOf, self.iri)
        )

    def _get_superclasses(self) -> Iterator[OntologyRelationship]:
        """Get all the superclasses of this relationship.

        Yields:
            The superrelationships.
        """
        yield self

        def closure(node, graph):
            yield from graph.objects(node, RDFS.subPropertyOf)

        yield from (
            self.session.from_identifier_typed(x, typing=OntologyRelationship)
            for x in self.session.graph.transitiveClosure(
                closure, self.identifier
            )
        )

        yield self.session.from_identifier_typed(
            OWL.topObjectProperty, typing=OntologyRelationship
        )

    def _get_subclasses(self) -> Iterator[OntologyRelationship]:
        """Get all the subclasses of this relationship.

        Yields:
            The subrelationships.
        """
        yield self

        def closure(node, graph):
            yield from graph.subjects(RDFS.subPropertyOf, node)

        if self.iri == OWL.topObjectProperty:
            yield from (
                self.session.from_identifier_typed(
                    s, typing=OntologyRelationship
                )
                for s in self.session.graph.subjects(
                    RDF.type, OWL.ObjectProperty
                )
            )
        else:
            yield from (
                self.session.from_identifier_typed(
                    x, typing=OntologyRelationship
                )
                for x in self.session.graph.transitiveClosure(
                    closure, self.identifier
                )
            )



"""A class defined in the ontology."""
from __future__ import annotations

import logging
from itertools import chain
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Set,
    Union,
)
from uuid import UUID

from rdflib import OWL, RDF, RDFS, BNode, URIRef
from rdflib.term import Identifier, Node

from simphony_osp.ontology.attribute import OntologyAttribute
from simphony_osp.ontology.entity import OntologyEntity
from simphony_osp.ontology.restriction import Restriction
from simphony_osp.utils.cache import lru_cache_timestamp
from simphony_osp.utils.datatypes import UID, AttributeValue, Triple

if TYPE_CHECKING:
    from simphony_osp.session.session import Session

logger = logging.getLogger(__name__)

BLACKLIST = {OWL.Nothing, OWL.Thing, OWL.NamedIndividual}

# The RDFLib namespace object is not as fast as it should be, so it is
# useful to reuse some IRIs throughout the file.
# TODO: Send PR to RDFLib to fix this upstream.
RDF_type = RDF.type
RDFS_domain = RDFS.domain
RDFS_subClassOf = RDFS.subClassOf
OWL_DatatypeProperty = OWL.DatatypeProperty
OWL_Restriction = OWL.Restriction
OWL_Thing = OWL.Thing
OWL_allValuesFrom = OWL.allValuesFrom
OWL_bottomDataProperty = OWL.bottomDataProperty
OWL_cardinality = OWL.cardinality
OWL_minCardinality = OWL.minCardinality
OWL_hasValue = OWL.hasValue
OWL_onProperty = OWL.onProperty
OWL_someValuesFrom = OWL.someValuesFrom
OWL_topDataProperty = OWL.topDataProperty


class OntologyClass(OntologyEntity):
    """A class defined in the ontology."""

    rdf_type = {OWL.Class, RDFS.Class}
    rdf_identifier = URIRef

    # ↓ --------------------- Public API --------------------- ↓ #

    @property
    @lru_cache_timestamp(lambda self: self.session.entity_cache_timestamp)
    def attributes(
        self,
    ) -> Mapping[OntologyAttribute, FrozenSet[Optional[AttributeValue]]]:
        """Get the attributes of this class.

        The attributes that all instances of this class are expected to
        have. A class can have attributes because one of its superclasses (
        including itself) has a default value for an attribute, or because
        the axioms affecting the superclass explicitly state that the class has
        such an attribute.
        """
        attributes: Dict[
            OntologyAttribute, FrozenSet[Optional[AttributeValue]]
        ]
        superclass: OntologyClass

        attributes = self._direct_attributes

        # Inherit attributes from the direct superclasses, which recursively
        # inherit from their own superclasses.
        for superclass in self.direct_superclasses:
            for attribute, default in superclass.attributes.items():
                attributes[attribute] = (
                    attributes.get(attribute, None) or default
                )
        return MappingProxyType(attributes)

    @property
    def axioms(self) -> FrozenSet[Restriction]:
        """Get all the axioms for the ontology class.

        Includes axioms inherited from its superclasses.

        Returns:
            Axioms for the ontology class.
        """
        axioms: Set[Restriction] = set()
        for superclass in self.superclasses:
            axioms |= self._compute_axioms(
                superclass.identifier, RDFS.subClassOf
            )
            axioms |= self._compute_axioms(
                superclass.identifier, OWL.equivalentClass
            )
        return frozenset(axioms)

    def __call__(
        self,
        session=None,
        iri: Optional[Union[URIRef, str, UID]] = None,
        uid: Optional[Union[UID, UUID, str, Node, int, bytes]] = None,
        _force: bool = False,
        **kwargs,
    ):
        """Create an OntologyIndividual object from this ontology class.

        Args:
            uid: The identifier of the Cuds object. Should be set to None in
                most cases. Then a new identifier is generated, defaults to
                None. Defaults to None.
            iri: The same as the uid, but exclusively for IRI identifiers.
            session (Session, optional): The session to create the cuds object
                in, defaults to None. Defaults to None.
            _force: Skip validity checks. Defaults to False.

        Raises:
            TypeError: Error occurred during instantiation.

        Returns:
            Cuds, The created cuds object
        """
        # TODO: Create ontology individuals, NOT CUDS objects.
        if None not in (uid, iri):
            raise ValueError(
                "Tried to initialize an ontology individual, both its IRI and "
                "UID. An ontology individual is constrained to have just one "
                "UID."
            )
        elif uid is not None and not isinstance(
            uid, (UID, UUID, str, Node, int, bytes)
        ):
            raise ValueError(
                "Provide an object of one of the following types as UID: "
                + ",".join(str(x) for x in (UID, UUID, str, Node, int, bytes))
            )
        elif iri is not None and not isinstance(iri, (URIRef, str, UID)):
            raise ValueError(
                "Provide an object of one of the following types as IRI: "
                + ",".join(str(x) for x in (URIRef, str, UID))
            )
        else:
            uid = (
                (UID(uid) if uid else None)
                or (UID(iri) if iri else None)
                or UID()
            )

        from simphony_osp.ontology.individual import OntologyIndividual

        # build attributes dictionary by combining
        # kwargs and defaults
        return OntologyIndividual(
            uid=uid,
            session=session,
            class_=self,
            attributes=self._kwargs_to_attributes(kwargs, _skip_checks=_force),
        )

    # ↑ --------------------- Public API --------------------- ↑ #

    def __init__(
        self,
        uid: UID,
        session: Optional[Session] = None,
        triples: Optional[Iterable[Triple]] = None,
        merge: bool = False,
    ) -> None:
        """Initialize the ontology class.

        Args:
            uid: UID identifying the ontology class.
            session: Session where the entity is stored.
            triples: Construct the class with the provided triples.
            merge: Whether overwrite the potentially existing entity in the
                session with the provided triples or just merge them with
                the existing ones.
        """
        super().__init__(uid, session, triples, merge=merge)

    @property
    @lru_cache_timestamp(lambda self: self.session.entity_cache_timestamp)
    def optional_attributes(self) -> FrozenSet[OntologyAttribute]:
        """Get the optional attributes of this class.

        The optional attributes are the non-mandatory attributes (those not
        returned by the `attributes` property) that have the class defined
        as their domain, or any of its superclasses.
        """
        superclass: OntologyClass
        attributes = frozenset(
            self._direct_optional_attributes
            | {
                attribute
                for superclass in self.direct_superclasses
                for attribute in superclass.optional_attributes
            }
        )
        return attributes

    @lru_cache_timestamp(lambda self: self.session.entity_cache_timestamp)
    def _compute_axioms(
        self, identifier: Identifier, predicate: URIRef
    ) -> FrozenSet[Restriction]:
        """Compute the axioms for the class with the given identifier.

        Does not include superclasses.

        Args:
            identifier: The IRI of the class.
            predicate: The predicate to which the class is connected to
                axioms (subclass or equivalentClass).

        Returns:
            Tuple of computed axioms.
        """
        axioms: Set[Restriction] = set()
        for o in self.session.graph.objects(identifier, predicate):
            if not isinstance(o, BNode):
                continue
            try:
                axioms.add(self.session.from_identifier_typed(o, Restriction))
            except (KeyError, TypeError):
                pass
        return frozenset(axioms)

    @property
    @lru_cache_timestamp(lambda self: self.session.entity_cache_timestamp)
    def _direct_attributes(
        self,
    ) -> Dict[OntologyAttribute, FrozenSet[Optional[AttributeValue]]]:
        """Get the non-inherited attributes of this ontology class.

        Returns:
            Mapping from attributes to the default value of the attribute.
            Mandatory attributes without a default value are mapped to `None`.
        """
        graph = self.session.graph
        attributes = dict()

        # From axioms.
        restrictions_on_data_properties = (
            # Yield both the restriction and the property.
            (restriction_iri, prop_iri)
            # Must be a restriction.
            for restriction_iri in graph.objects(
                self.identifier, RDFS_subClassOf
            )
            if (restriction_iri, RDF_type, OWL_Restriction) in graph
            # The property must be a DatatypeProperty.
            for prop_iri in (
                graph.value(restriction_iri, OWL_onProperty, any=False),
            )
            if (prop_iri, RDF_type, OWL_DatatypeProperty) in graph
        )
        for restriction_iri, prop_iri in restrictions_on_data_properties:
            attribute = self.session.from_identifier_typed(
                prop_iri, typing=OntologyAttribute
            )

            # Get restriction default.
            default = graph.value(restriction_iri, OWL_hasValue)

            # Determine if attribute is mandatory.
            obligatory = any(
                (
                    self.session.graph.value(
                        restriction_iri, OWL_someValuesFrom
                    ),
                    self.session.graph.value(
                        restriction_iri, OWL_allValuesFrom
                    ),
                    self.session.graph.value(restriction_iri, OWL_hasValue),
                    self.session.graph.value(restriction_iri, OWL_cardinality)
                    != 0,
                    self.session.graph.value(
                        restriction_iri, OWL_minCardinality != 0
                    ),
                )
            )

            if default or obligatory:
                attributes[attribute] = attributes.get(attribute, default)

        # TODO more cases
        return attributes

    @property
    @lru_cache_timestamp(lambda self: self.session.entity_cache_timestamp)
    def _direct_optional_attributes(
        self,
    ) -> FrozenSet[OntologyAttribute]:
        """Get the non-inherited optional attributes of this ontology class.

        The optional non-inherited attributes are the non-mandatory attributes
        (those not returned by the `_direct_attributes` property) that have
        the class defined as their domain, or any of its superclasses.

        Returns:
            A frozen set containing all the non-inherited optional
            attributes of the class.
        """
        graph = self.session.graph
        attributes = set()

        # Class is part of the domain of a DatatypeProperty.
        blacklist = [OWL_topDataProperty, OWL_bottomDataProperty]
        target_properties = (
            s
            for s in graph.subjects(RDFS_domain, self.identifier)
            if (s, RDF_type, OWL_DatatypeProperty) in graph or s in blacklist
        )
        for identifier in target_properties:
            attribute = self.session.from_identifier_typed(
                identifier, typing=OntologyAttribute
            )
            attributes.add(attribute)
        return frozenset(attributes)

    def _get_direct_superclasses(self) -> Iterator[OntologyClass]:
        """Get all the direct superclasses of this ontology class.

        Returns:
            The direct superclasses.
        """
        for o in self.session.graph.objects(self.iri, RDFS_subClassOf):
            try:
                yield self.session.from_identifier_typed(
                    o, typing=OntologyClass
                )
            except TypeError:
                pass

    def _get_direct_subclasses(self) -> Iterator[OntologyClass]:
        """Get all the direct subclasses of this ontology class.

        Returns:
            The direct subclasses.
        """
        for s in self.session.graph.subjects(RDFS_subClassOf, self.iri):
            try:
                yield self.session.from_identifier_typed(
                    s, typing=OntologyClass
                )
            except TypeError:
                pass

    def _get_superclasses(self) -> Iterator[OntologyClass]:
        """Get all the superclasses of this ontology class.

        Yields:
            The superclasses.
        """
        yield self

        def closure(node, graph):
            yield from graph.objects(node, RDFS_subClassOf)

        yield from filter(
            lambda x: isinstance(x, OntologyClass),
            (
                self.session.from_identifier(x)
                for x in self.session.graph.transitiveClosure(
                    closure, self.identifier
                )
            ),
        )

        yield self.session.from_identifier(OWL_Thing)

    def _get_subclasses(self) -> Iterator[OntologyClass]:
        """Get all the subclasses of this ontology class.

        Yields:
            The subclasses.
        """
        yield self

        def closure(node, graph):
            yield from graph.subjects(RDFS_subClassOf, node)

        yield from filter(
            lambda x: isinstance(x, OntologyClass),
            (
                self.session.from_identifier(x)
                for x in self.session.graph.transitiveClosure(
                    closure, self.identifier
                )
            ),
        )

    def _kwargs_to_attributes(
        self, kwargs: Mapping, _skip_checks: bool
    ) -> Mapping[OntologyAttribute, FrozenSet[Any]]:
        """Combine class attributes with the ones from the given kwargs.

        Args:
            kwargs: The user specified keyword arguments.
            _skip_checks: When true, allow mandatory attributes to be left
                undefined.

        Raises:
            TypeError: Unexpected keyword argument.
            TypeError: Missing keyword argument.

        Returns:
            The resulting mixture.
        """
        kwargs = dict(kwargs)

        attributes = dict()
        attribute_declaration = {
            attribute: (default, mandatory)
            for attribute, default, mandatory in chain(
                ((attr, None, False) for attr in self.optional_attributes),
                ((attr, def_, True) for attr, def_ in self.attributes.items()),
            )
        }
        for attribute, (
            default,
            obligatory,
        ) in attribute_declaration.items():
            labels = set(attribute.iter_labels(return_literal=False))
            if attribute.namespace is not None:
                labels |= {
                    attribute.identifier[len(attribute.namespace.iri) :]
                }
            label = next(filter(lambda x: x in kwargs, labels), None)
            if label is not None:
                attributes[attribute] = kwargs[label]
                del kwargs[label]
            elif not _skip_checks and obligatory:
                raise TypeError(
                    "Missing keyword argument: %s" % attribute.label
                )
            elif default is not None:
                attributes[attribute] = default
            else:
                continue

            # Turn attribute into a mutable sequence.
            if not isinstance(attributes[attribute], FrozenSet):
                attributes[attribute] = [attributes[attribute]]
            else:
                attributes[attribute] = list(attributes[attribute])

            # Set the appropriate hashable data type for the arguments.
            attributes[attribute] = frozenset(
                {
                    attribute.convert_to_datatype(value)
                    for value in attributes[attribute]
                }
            )

        # Check validity of arguments
        if not _skip_checks and kwargs:
            raise TypeError("Unexpected keyword arguments: %s" % kwargs.keys())
        return MappingProxyType(attributes)



"""Generic ontology parser abstract class.

Also contains a `Parser` class for backwards compatibility.
"""
from __future__ import annotations

import logging
import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, Set, Union

import yaml
from rdflib import Graph, URIRef

logging = logging.getLogger(__name__)


class OntologyParser(ABC):
    """Generic ontology parser abstract class."""

    graph: Graph

    @abstractmethod
    def __init__(self):
        """Initialize the parser.

        NOTE: It is recommended to do it in such a way that the property
        `graph` is lazily loaded, since its computation can take a significant
        amount of time.
        """
        pass

    @property
    @abstractmethod
    def identifier(self) -> str:
        """Get the identifier of the loaded ontology package.

        Returns:
            str: The identifier of the loaded ontology package.
        """
        pass

    @property
    @abstractmethod
    def namespaces(self) -> Dict[str, URIRef]:
        """Fetch the namespaces from the ontology files.

        Returns:
            Dict[str, URIRef]: A dictionary containing the defined namespace
                names and URIs.
        """
        pass

    @property
    @abstractmethod
    def requirements(self) -> Set[str]:
        """Fetch the requirements from the ontology file."""
        pass

    @abstractmethod
    def install(self, destination: Union[str, Path]) -> None:
        """Store the parsed files at the given destination.

        This function is meant to copy the ontology to the SimPhoNy data
        directory. So usually the destination will be `~/.simphony-osp`.

        Args:
            destination: the SimPhoNy data directory.
        """
        pass

    @staticmethod
    def parse_file_path(file_identifier: Union[str, Path]) -> str:
        """Get the correct file path for a given identifier.

        For a given one, i.e. translate non
        paths to osp/core/ontology/files/*.yml

        Args:
            file_identifier: A filepath or file identifier.

        Returns:
            The translated file path.
        """
        file_identifier = Path(file_identifier)
        if str(file_identifier).endswith(".yml"):
            return str(file_identifier)
        file_identifier = str(file_identifier).lower()
        a = os.path.join(
            os.path.dirname(__file__),
            "../files",
            f"{file_identifier}.ontology.yml",
        )
        b = os.path.join(
            os.path.dirname(__file__), "../files", f"{file_identifier}.yml"
        )
        if os.path.exists(a):
            return os.path.abspath(a)
        return os.path.abspath(b)

    @staticmethod
    def load_yaml(path: Union[str, Path]):
        """Load the given YAML file.

        Args:
            path: the path of the YAML file.
        """
        with open(path, "r") as file:
            doc = yaml.safe_load(file)
            if doc is None:
                raise SyntaxError(f"File {path} is empty.")
        return doc

    @staticmethod
    def is_owl_ontology(doc: dict) -> bool:
        """Tells whether a given YAML doc is an OWL ontology config file.

        Args:
            doc: the doc obtained after reading the YAML config file.
        """
        import simphony_osp.ontology.parser.owl.keywords as keywords

        return all(
            x in doc for x in (keywords.RDF_FILE_KEY, keywords.IDENTIFIER_KEY)
        )

    @classmethod
    def get_parser(cls, path: Union[str, Path]) -> OntologyParser:
        """Parse the given YAML files.

        Args:
            path: path to the YAML file
        """
        from simphony_osp.ontology.parser.owl.parser import OWLParser

        file_path = cls.parse_file_path(path)
        yaml_doc = cls.load_yaml(file_path)
        if cls.is_owl_ontology(yaml_doc):
            parser = OWLParser(file_path)
        else:
            raise SyntaxError(f"Invalid format of file {file_path}")
        return parser



"""Pico is a tool to manage SimPhoNy's installed ontologies.

This file contains the non-public-facing elements of pico.
"""

import glob
import logging
import os
from pathlib import Path
from typing import (
    Callable,
    Dict,
    Hashable,
    Iterable,
    Iterator,
    List,
    MutableMapping,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from simphony_osp.ontology.namespace import OntologyNamespace
from simphony_osp.ontology.parser import OntologyParser
from simphony_osp.session.session import Session

logger = logging.getLogger(__name__)

HASHABLE = TypeVar("HASHABLE", bound=Hashable)


def graph_set_to_dict(
    arcs: Set[Tuple[HASHABLE, HASHABLE]]
) -> MutableMapping[HASHABLE, Set[HASHABLE]]:
    """Convert the graph from a set of arcs to a dictionary (fast lookup).

    Transforms a set of arcs into a dictionary of arcs.

    Args:
        arcs: A set of arcs (directed edges) represented by tuples, where
            their first element is the tail and the second the head.

    Returns:
        A dictionary of arcs, where the keys represent the tail of an arc
        and each value is a set of heads for such tail. Therefore,
        each key-value pair represents several arcs.
    """
    graph = dict()
    for x, y in arcs:
        if x not in graph:
            graph[x] = {y}
        else:
            graph[x] |= {y}
        if y not in graph:
            graph[y] = set()
    return graph


def depth_first_search(
    graph: MutableMapping[HASHABLE, Set[HASHABLE]],
    start: Iterable[HASHABLE],
    goal: Callable[[HASHABLE], bool],
) -> Iterator[HASHABLE]:
    """Implementation of depth first search graph algorithm.

    This implementation assumes that the provided set of arcs represents a
    directed graph. Arcs will only be traversed from their tails to their
    heads.

    This implementation is in the form of a generator. This means that the
    algorithm will look for a "goal" node, but instead of returning it,
    it will yield it. This implies that after returning the first node,
    if the iterator is called again, it will return more "goal" nodes that
    satisfy the goal function.

    Args:
        graph: A dictionary of arcs, where the keys represent the tail of an
            arc and each value is a set of heads for such tail. Therefore, each
            key-value pair represents several arcs.
        start: The starting nodes.
        goal: A function that evaluates at every node to determine whether
            it is a "goal" node (should be yielded) or not.

    Yields:
        Nodes that satisfy the goal condition.
    """
    visited = set()
    queue = list(set(start))
    while queue:
        node = queue.pop()
        if goal(node):
            yield node
        visited.add(node)
        for neighbor in graph.get(node, set()):
            if neighbor not in visited:
                queue.append(neighbor)


def topological_sort(graph: MutableMapping[Hashable, Set[Hashable]]):
    """Kanh's algorithm for topological sort.

    Kahn, Arthur B. (1962), "Topological sorting of large networks",
    Communications of the ACM, 5 (11): 558–562, doi:10.1145/368996.369025,
    S2CID 16728233

    Args:
        graph: A dictionary of arcs, where the keys represent the tail of an
            arc and each value is a set of heads for such tail. Therefore, each
            key-value pair represents several arcs.
    """
    result = []
    no_incoming_edges = set(graph.keys()) - {
        x for s in graph.values() for x in s
    }
    while no_incoming_edges:
        node = no_incoming_edges.pop()
        result += [node]
        for m in set(graph[node]):
            graph[node].remove(m)
            if m not in {x for s in graph.values() for x in s}:
                no_incoming_edges.add(m)

    if {y for x in graph.values() for y in x}:
        raise ValueError(
            "The provided set of arcs has cycles, therefore "
            "topological sorting is unfeasible."
        )
    return tuple(result)


class Pico:
    """This class manages the installed ontologies."""

    def __init__(self, path: Optional[Union[str, Path]] = None):
        """Initialize pico.

        Args:
            path: Installation destination. When unspecified, the default
                location is used.
        """
        self._ontology = Session(identifier="pico ontologies", ontology=True)
        os.makedirs(self.path, exist_ok=True)
        if path is not None:
            self.path = path
        self._reload_installed_ontologies()

    @property
    def ontology(self) -> Session:
        """Ontology containing all the installed packages.

        This ontology is reloaded every time a change in the installed
        ontologies is done through an instance of this class.

        WARNING: Be careful not to instantiate two instances of `Pico` using
        the same `path`.
        """
        return self._ontology

    @property
    def path(self) -> Path:
        """Path where ontologies are installed."""
        return self._path or self.get_default_installation_path()

    @path.setter
    def path(self, value: Optional[str]):
        """Set the path where ontologies are installed."""
        self._path = value if value is None else Path(value)
        self._reload_installed_ontologies()

    @property
    def packages(self) -> Tuple[str]:
        """Returns the identifiers of all installed packages."""
        return tuple(self._package_names)

    @property
    def namespaces(self) -> Tuple[OntologyNamespace]:
        """Returns namespace objects for all the installed namespaces."""
        return tuple(self.ontology.namespaces)

    def install(
        self, *files: Union[Path, str], overwrite: bool = False
    ) -> None:
        """Install ontology packages.

        Args:
            files: Paths of `yml` files describing the ontologies to install.
                Alternatively, identifiers of ontology packages that are
                bundled with SimPhoNy.
            overwrite: Whether to overwrite already installed ontology
                packages.
        """
        files: Tuple[Path] = tuple(Path(file) for file in files)
        installed_identifiers: Set[str] = set(self._package_names)
        new_parsers: Dict[str, Tuple[OntologyParser, Path]] = {
            parser.identifier: (parser, path)
            for path in files
            for parser in (OntologyParser.get_parser(path),)
        }
        new_identifiers = set(new_parsers)
        if overwrite:
            installed_identifiers -= new_identifiers
        skip_identifiers = new_identifiers & installed_identifiers
        install_parsers: Dict[str, Tuple[OntologyParser, Path]] = {
            identifier: (parser, path)
            for identifier, (parser, path) in new_parsers.items()
            if identifier not in skip_identifiers
        }
        for identifier in skip_identifiers:
            logger.info(
                "Skipping package with identifier %s, "
                "because it is already installed." % identifier
            )
        del new_parsers, new_identifiers, skip_identifiers

        # Resolve the dependencies using DFS with the goal function below.
        # - create requirements graph
        requirements_graph = {
            identifier: parser.requirements
            for identifier, (parser, path) in install_parsers.items()
        }
        missing_requirements = set()

        def explore_dependencies(node: str) -> False:
            """Generate the dependency graph.

            Goal function for the DFS algorithm that loads each
            package (if not loaded already or installed), evaluates its
            dependencies and adds them to the graph.

            When a package fails to load, it is added to a set of missing
            requirements, that is later used to raise an exception.
            """
            if (
                node not in installed_identifiers
                and node not in install_parsers
            ):
                try:
                    parser = OntologyParser.get_parser(node)
                    install_parsers[parser.identifier] = (parser, Path(node))
                    requirements_graph[node] = parser.requirements
                except FileNotFoundError:
                    requirements_graph[node] = set()
                    missing_requirements.add(node)
            return False

        depth_first_search_iterator = depth_first_search(
            requirements_graph, requirements_graph, explore_dependencies
        )
        # Since the goal function is always false, asking for the next item
        # is enough to explore the whole graph.
        next(depth_first_search_iterator, None)

        # Raise exception if there are missing requirements.
        if missing_requirements:
            missing_requirements_dict = dict()
            print(install_parsers)
            for identifier, (parser, path) in install_parsers.items():
                missing = tuple(
                    x for x in parser.requirements if x in missing_requirements
                )
                if missing:
                    missing_requirements_dict[parser.identifier] = missing
            raise RuntimeError(
                "Installation failed. Unsatisfied requirements: \n - %s"
                % "\n - ".join(
                    [f"{n}: {r}" for n, r in missing_requirements_dict.items()]
                )
            )

        if install_parsers:
            logger.info(
                "Will install the following packages: %s."
                % ", ".join(install_parsers)
            )
            logger.info(
                "Will install the following namespaces: %s."
                % ", ".join(
                    name
                    for identifier, (parser, path) in install_parsers.items()
                    for name in parser.namespaces.keys()
                )
            )

            for identifier, (parser, path) in install_parsers.items():
                parser.install(self.path)

            logger.info("Installation successful")

            self._reload_installed_ontologies()

    def uninstall(self, *packages: str) -> None:
        """Uninstall ontology packages.

        Args:
            packages: Identifiers of the ontology packages to uninstall.
        """
        installed_identifiers: Set[str] = set(self._package_names)
        remove_identifiers: Set[str] = set(packages)
        not_installed_identifiers: Set[str] = (
            remove_identifiers - installed_identifiers
        )
        if not_installed_identifiers:
            raise RuntimeError(
                f"Cannot uninstall the following packages, "
                f"as they are not installed: "
                f"{', '.join(not_installed_identifiers)}."
            )

        # Load the installed packages and create a dependency graph. In such
        # a graph, given an arc, the tail is a package and the head a
        # package that depends on it.
        paths = {
            os.path.join(self.path, yml)
            for yml in (
                x
                for x in os.listdir(self.path)
                if "yml" in str(os.path.splitext(x)[1])
            )
        }
        parsers = {OntologyParser.get_parser(path) for path in paths}
        directed_edges = {
            (requirement, parser.identifier)
            for parser in parsers
            for requirement in parser.requirements
        }
        requirements_graph = graph_set_to_dict(directed_edges)
        del directed_edges, parsers, paths

        depth_first_search_iterator = depth_first_search(
            requirements_graph,
            remove_identifiers,
            lambda node: True,
            # Every node reachable from the starting nodes is required to be
            # removed to leave the system in a healthy state.
        )
        should_be_removed = set(depth_first_search_iterator)

        # Block package removal if another package depends on it.
        conflicts = should_be_removed - remove_identifiers
        if conflicts:
            # Create a graph of conflicts from the requirements graph.
            detailed_conflicts = {
                (target, source)
                for source, targets in requirements_graph.items()
                if source in should_be_removed
                for target in targets
                if target in conflicts
            }
            detailed_conflicts = graph_set_to_dict(detailed_conflicts)
            # Filter out tails with no heads.
            detailed_conflicts = {
                key: value
                for key, value in detailed_conflicts.items()
                if value
            }

            message = (
                "Cannot remove package{plural} {cannot_remove}{comma} "
                "because other installed packages depend on {pronoun}: "
                "{dependency_list}. "
                "Please remove the packages {all_packages_to_remove} "
                "all together."
            )
            plural = "s" if len(remove_identifiers) > 1 else ""
            comma = ";" if plural else ","
            pronoun = "them" if plural else "it"
            cannot_remove = ", ".join(remove_identifiers)
            one_dependency = next(iter(detailed_conflicts.values()))
            all_dependencies_equal = all(
                one_dependency == x for x in detailed_conflicts.values()
            )
            if all_dependencies_equal:
                dependency_list = ", ".join(detailed_conflicts)
            else:
                dependency_list = set()
                for package, conflict_list in detailed_conflicts.items():
                    dependency_list.add(
                        f"package {package} depends on "
                        f"{', '.join(conflict_list)}"
                    )
                dependency_list = "; ".join(dependency_list)
            all_packages_to_remove = set(should_be_removed)
            last_package = all_packages_to_remove.pop()
            all_packages_to_remove = (
                ", ".join(all_packages_to_remove) + " and " + last_package
            )
            message = message.format(
                plural=plural,
                cannot_remove=cannot_remove,
                comma=comma,
                pronoun=pronoun,
                dependency_list=dependency_list,
                all_packages_to_remove=all_packages_to_remove,
            )
            raise RuntimeError(message)

        if remove_identifiers:
            files_to_remove = (
                glob.glob(os.path.join(self.path, f"{identifier}*"))
                for identifier in remove_identifiers
            )
            files_to_remove = {
                file for glob_list in files_to_remove for file in glob_list
            }
            for file in files_to_remove:
                os.remove(file)

            logger.info("Uninstallation successful")

            self._reload_installed_ontologies()

    @staticmethod
    def get_default_installation_path() -> Path:
        """Get the path where ontologies are installed by default."""
        # Get the path from the environment variable if defined.
        path = os.environ.get("SIMPHONY_ONTOLOGIES_DIR")
        if path:
            path = Path(path)

        # If the environment variable is not defined, use the default path.
        path = path or Path.home() / ".simphony-osp" / "ontologies"

        return path

    def set_default_installation_path(
        self, value: Optional[Union[str, Path]]
    ) -> None:
        """Set the path where ontologies are installed by default."""
        path = self.path

        if value is not None:
            os.environ["SIMPHONY_ONTOLOGIES_DIR"] = str(value)
        else:
            del os.environ["SIMPHONY_ONTOLOGIES_DIR"]

        if path != self.path:
            self._reload_installed_ontologies()

    _path: Optional[Path] = None
    _ontology: Session

    @property
    def _package_names(self) -> Iterator[str]:
        """Get the names of the installed packages.

        Returns:
            Identifiers of the installed packages.
        """
        return (
            OntologyParser.get_parser(path).identifier
            for path in self._package_paths
        )

    @property
    def _package_paths(self) -> Iterator[Path]:
        """Get the paths of the files describing the installed packages."""
        return (
            Path(self.path) / yml
            for yml in (
                str(x)
                for x in os.listdir(self.path)
                if "yml" in str(os.path.splitext(x)[1])
            )
        )

    @property
    def _package_names_paths(self) -> Iterator[Tuple[str, Path]]:
        """Get both the identifiers and the paths of the installed packages."""
        return (
            (OntologyParser.get_parser(path).identifier, path)
            for path in self._package_paths
        )

    @property
    def _topologically_sorted_parsers(self) -> List[OntologyParser]:
        """Returns a list of parsers, topologically sorted.

        As the parsers are topologically sorted with respect to their
        requirements, they can be loaded in order without triggering missing
        ontology entity errors.
        """
        paths = self._package_paths
        parsers = {OntologyParser.get_parser(path) for path in paths}
        directed_edges = {
            (requirement, parser.identifier)
            for parser in parsers
            for requirement in parser.requirements or (None,)
        }
        graph = graph_set_to_dict(directed_edges)
        sorted_identifiers = topological_sort(graph)
        parsers = sorted(
            parsers, key=lambda x: sorted_identifiers.index(x.identifier)
        )
        return parsers

    def _reload_installed_ontologies(self):
        self._ontology.clear(force=True)
        for parser in self._topologically_sorted_parsers:
            self._ontology.load_parser(parser)


# Create a pico singleton for the default directory and set the installed
# packages as the default ontology.
pico = Pico()
Session.default_ontology = pico.ontology



"""Implementation of a remote RDFLib store over websockets."""

import json
import os
import tempfile
from itertools import chain
from typing import (
    Any,
    BinaryIO,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Union,
)

from rdflib import Graph, URIRef
from rdflib.plugins.parsers.jsonld import to_rdf as json_to_rdf
from rdflib.plugins.stores.memory import SimpleMemory
from rdflib.store import Store

from simphony_osp.interfaces.interface import BufferType, Interface
from simphony_osp.interfaces.remote.common import COMMAND, parse_uri
from simphony_osp.interfaces.remote.engine import CommunicationEngineClient
from simphony_osp.utils.datatypes import Pattern, Triple


class RemoteInterface(Interface):
    """Interface that communicates with a remote interface server."""

    # Connection and authentication.
    _uri: Optional[str] = None
    _username: Optional[str] = None
    _password: Optional[str] = None
    _engine: Optional[CommunicationEngineClient] = None

    @staticmethod
    def _handle_response(
        data: str, files: List[BinaryIO]
    ) -> (Dict[str, Any], List[BinaryIO]):
        data = json.loads(data or "{}")
        return data, files

    def __getattribute__(self, item):
        """Check whether the interface on the remote side has the attribute."""
        if item in (
            "compute",
            "add",
            "remove",
            "triples",
            "save",
            "load",
            "delete",
            "hash",
            "rename",
        ):
            response, _ = self._engine.send(
                COMMAND.HASATTR, json.dumps({"item": item})
            )
            if response[COMMAND.HASATTR] is False:
                raise AttributeError(item)
        return super().__getattribute__(item)

    # Interface
    # ↓ ----- ↓

    entity_tracking = False

    cache = True

    def open(self, configuration: str, create: bool = False) -> None:
        """Implements the OPEN command."""
        if self._engine is not None:
            raise RuntimeError(
                f"Already connected to {self._uri}. Please "
                f"close the session first."
            )

        # Decode configuration string.
        uri, username, password = (
            parse_uri(configuration) if configuration else (None,) * 3
        )
        self._uri, self._username, self._password = uri, username, password

        self._engine = CommunicationEngineClient(
            uri=self._uri,
            handle_response=self._handle_response,
        )

        # Send authentication command
        response, _ = self._engine.send(
            COMMAND.AUTHENTICATE,
            json.dumps(
                {"username": self._username, "password": self._password}
            ),
        )

        if self.base is None:
            remote_store = RemoteStoreClient(engine=self._engine)
            self.base = Graph(store=remote_store)

        # response, _ = self._engine.send(
        #     COMMAND.OPEN,
        #     json.dumps({
        #         "configuration": configuration,
        #         "create": create
        #     }),
        # )
        # return response[COMMAND.OPEN]

    def close(self) -> None:
        """Implements the CLOSE command."""
        if self._engine is None:
            return

        # response, _ = self._engine.send(
        #     COMMAND.CLOSE,
        #     json.dumps({}),
        # )

        # Close connection and clear connection information.
        self._engine.close()
        self._uri, self._username, self._password = (None,) * 3

        # return response[COMMAND.CLOSE]

    def populate(self) -> None:
        """Implements the POPULATE command."""
        response, _ = self._engine.send(
            COMMAND.POPULATE,
            json.dumps({}),
        )
        return response[COMMAND.POPULATE]

    def commit(self) -> None:
        """Implements the ROOT command."""
        response, _ = self._engine.send(
            COMMAND.COMMIT,
            json.dumps({}),
        )
        return response[COMMAND.COMMIT]

    def compute(
        self,
        **kwargs: Union[
            str,
            int,
            float,
            bool,
            None,
            Iterable[Union[str, int, float, bool, None]],
        ],
    ) -> None:
        """Implements the COMPUTE command."""
        response, _ = self._engine.send(
            COMMAND.COMPUTE,
            json.dumps({"kwargs": kwargs}),
        )
        return response.get(COMMAND.COMPUTE)

    def add(self, triple: Triple) -> bool:
        """Implements the ADD command."""
        g = Graph()
        g.add(triple)
        response, _ = self._engine.send(
            COMMAND.ADD,
            json.dumps(g.serialize(format="json-ld")),
        )
        return response.get(COMMAND.ADD)

    def remove(self, pattern: Triple) -> Iterator[Triple]:
        """Implements the REMOVE command."""
        g = Graph()
        s = pattern[0] if pattern[0] is not None else URIRef("none:None")
        p = pattern[1] if pattern[1] is not None else URIRef("none:None")
        o = pattern[2] if pattern[2] is not None else URIRef("none:None")
        g.add((s, p, o))
        response, _ = self._engine.send(
            COMMAND.REMOVE,
            g.serialize(format="turtle"),
        )
        yield from json_to_rdf(response[COMMAND.REMOVE], Graph())

    def triples(self, pattern: Triple) -> Iterator[Triple]:
        """Implements the TRIPLES command."""
        g = Graph()
        s = pattern[0] if pattern[0] is not None else URIRef("none:None")
        p = pattern[1] if pattern[1] is not None else URIRef("none:None")
        o = pattern[2] if pattern[2] is not None else URIRef("none:None")
        g.add((s, p, o))
        response, _ = self._engine.send(
            COMMAND.TRIPLES,
            g.serialize(format="turtle"),
        )
        yield from json_to_rdf(response[COMMAND.TRIPLES], Graph())

    def save(self, key: str, file: BinaryIO) -> None:
        """Implements the SAVE command."""
        file = tempfile.NamedTemporaryFile(delete=False)
        try:
            response, _ = self._engine.send(
                COMMAND.COMPUTE,
                json.dumps(
                    {
                        "key": key,
                    }
                ),
                [file.name],
            )
        finally:
            os.remove(file.name)
        return response.get(COMMAND.SAVE)

    def load(self, key: str) -> BinaryIO:
        """Implements the LOAD command."""
        response, files = self._engine.send(
            COMMAND.LOAD,
            json.dumps(
                {
                    "key": key,
                }
            ),
        )
        return files

    def delete(self, key: str) -> BinaryIO:
        """Implements the DELETE command."""
        response, _ = self._engine.send(
            COMMAND.DELETE,
            json.dumps(
                {
                    "key": key,
                }
            ),
        )
        return response[COMMAND.DELETE]

    def hash(self, key: str) -> str:
        """Implements the HASH command."""
        response, _ = self._engine.send(
            COMMAND.HASH,
            json.dumps(
                {
                    "key": key,
                }
            ),
        )
        return response[COMMAND.HASH]

    def rename(self, key: str, new_key: str) -> None:
        """Implements the RENAME command."""
        response, _ = self._engine.send(
            COMMAND.RENAME,
            json.dumps(
                {
                    "key": key,
                }
            ),
        )
        return response[COMMAND.RENAME]

    # Interface
    # ↑ ----- ↑


class RemoteStoreClient(Store):
    """Implementation of a remote RDFLib store over websockets.

    It consists of a client and a server. The triples are buffered until
    `commit` is called.
    """

    # Connection and authentication.
    _engine: Optional[CommunicationEngineClient] = None

    # RDFLib
    # ↓ -- ↓

    transaction_aware = True
    graph_aware = False
    context_aware = False
    formula_aware = False

    def __init__(
        self,
        configuration: Optional[str] = None,
        identifier: Optional[URIRef] = None,
        engine: CommunicationEngineClient = None,
    ) -> None:
        """Construct the remote store client."""
        if engine is None:
            raise ValueError("No engine provided.")
        self._engine = engine
        self._reset_buffers()  # Creates the buffers for the first time.
        super().__init__(configuration, identifier)

    def open(self, configuration: str, create: bool = False) -> None:
        """Initializes the communications engine client.

        Args:
            configuration: WebSocket URI.
            create: Currently not used, just to comply with RDFLib's API.
        """
        self._reset_buffers()
        response, _ = self._engine.send(
            COMMAND.STORE_OPEN,
            json.dumps({"configuration": configuration, "create": create}),
        )
        return response[COMMAND.STORE_OPEN]

    def close(self, commit_pending_transaction: bool = False) -> None:
        """Close the connection to the remote store.

        Args:
            commit_pending_transaction: Whether to commit any pending
                transactions before disconnecting.
        """
        response, _ = self._engine.send(
            COMMAND.STORE_CLOSE,
            json.dumps(
                {"commit_pending_transaction": commit_pending_transaction}
            ),
        )
        return response[COMMAND.STORE_CLOSE]

    def add(self, triple: Triple, context, quoted=False) -> None:
        """Adds triples to the store.

        Since the actual addition happens during a commit, this method just
        buffers the changes.
        """
        self._buffers[BufferType.DELETED].remove(triple)
        self._buffers[BufferType.ADDED].add(triple)

    def remove(self, triple_pattern: Pattern, context=None) -> None:
        """Remove triples from the store.

        Since the actual removal happens during a commit, this method just
        buffers the changes.
        """
        self._buffers[BufferType.ADDED].remove(triple_pattern)
        existing_triples_to_delete = (
            triple for triple in self._remote_triples(triple_pattern)
        )
        for triple in existing_triples_to_delete:
            self._buffers[BufferType.DELETED].add(triple)

    def triples(
        self, triple_pattern: Pattern, context=None
    ) -> Iterator[Triple]:
        """Query triples patterns.

        Merges the buffered changes with the data stored on the remote
        store. Therefore, this method involves network traffic.
        """
        # Existing minus added and deleted triples.
        triple_pool = filter(
            lambda x: x
            not in chain(
                self._buffers[BufferType.DELETED].triples(triple_pattern),
                self._buffers[BufferType.ADDED].triples(triple_pattern),
            ),
            self._remote_triples(triple_pattern),
        )
        # Include added triples (previously they were excluded in order not
        # to duplicate triples).
        triple_pool = chain(
            triple_pool,
            self._buffers[BufferType.ADDED].triples(triple_pattern),
        )
        for triple in triple_pool:
            yield triple, iter(())

    def __len__(self, context=None) -> int:
        """Get the number of triples in the store.

        For more details, check RDFLib's documentation.
        """
        i = 0
        for _ in self.triples((None, None, None)):
            i += 1
        return i

    def bind(self, prefix, namespace):
        """Bind a namespace to a prefix."""
        raise NotImplementedError

    def namespace(self, prefix):
        """Bind a namespace to a prefix."""
        raise NotImplementedError

    def prefix(self, prefix):
        """Get a bound namespace's prefix."""
        raise NotImplementedError

    def namespaces(self):
        """Get the bound namespaces."""
        raise NotImplementedError

    def query(self, *args, **kwargs):
        """Perform a SPARQL query on the store."""
        return super().query(*args, **kwargs)

    def update(self, *args, **kwargs):
        """Perform a SPARQL update query on the store."""
        return super().update(*args, **kwargs)

    def commit(self) -> None:
        """Commit buffered changes."""
        # data are the buffers, serialized (json-ld)
        self._engine.send(
            COMMAND.STORE_ADD,
            self._buffers[BufferType.ADDED].serialize(format="json-ld"),
        )
        self._engine.send(
            COMMAND.STORE_REMOVE,
            self._buffers[BufferType.DELETED].serialize(format="turtle"),
        )
        self._engine.send(COMMAND.STORE_COMMIT, "")

        self._reset_buffers()

    def rollback(self) -> None:
        """Discard uncommitted changes."""
        self._reset_buffers()

    # RDFLib
    # ↑ -- ↑

    def _remote_triples(self, triple_pattern: Pattern) -> Iterator[Triple]:
        """Fetch triples from the remote store.

        Args:
            triple_pattern: The triple pattern to query the remote store.
        """
        triple_pattern = tuple(
            x or URIRef("none:None") for x in triple_pattern
        )
        pattern = Graph()
        pattern.add(triple_pattern)
        response, _ = self._engine.send(
            COMMAND.STORE_TRIPLES, pattern.serialize(format="turtle")
        )
        yield from json_to_rdf(response[COMMAND.STORE_TRIPLES], Graph())

    def _reset_buffers(self) -> None:
        """Replaces the existing buffers by empty buffers."""
        self._buffers = {
            buffer_type: Graph(SimpleMemory()) for buffer_type in BufferType
        }



"""Utility functions for printing ontology entities objects in a nice way."""

import sys
from functools import reduce
from operator import add
from typing import Iterable, Optional, Set, Union

from simphony_osp.namespaces import owl
from simphony_osp.ontology.attribute import OntologyAttribute
from simphony_osp.ontology.composition import Composition
from simphony_osp.ontology.entity import OntologyEntity
from simphony_osp.ontology.individual import OntologyIndividual
from simphony_osp.ontology.oclass import OntologyClass
from simphony_osp.ontology.relationship import OntologyRelationship
from simphony_osp.ontology.restriction import Restriction


def pretty_print(
    entity: OntologyEntity,
    rel: Union[
        OntologyRelationship, Iterable[OntologyRelationship]
    ] = owl.topObjectProperty,
    file=sys.stdout,
):
    """Print the given ontology entity in a human-readable way.

    The identifier, the type, the ancestors and the content are printed.

    Args:
        entity: Entity to be printed.
        file: The file to print to.
        rel: The relationships to consider when searching for sub-elements.
    """
    # Fix the order of each element by pre-populating a dictionary.
    pp = {
        x: ""
        for x in (
            "title",
            "identifier",
            "classes",
            "superclasses",
            "attributes",
            "subelements",
        )
    }

    # Title
    pp["title"] = f"{_pp_entity_name(entity)}:"

    # Superclasses
    superclasses = (
        {
            superclass
            for class_ in entity.classes
            for superclass in class_.superclasses
        }
        if isinstance(entity, OntologyIndividual)
        else set(entity.superclasses)
    )
    labels = _pp_list_of_labels_or_uids(superclasses, namespace=True)
    pp["superclasses"] = "\n  superclasses: " + labels

    if isinstance(entity, OntologyIndividual):
        # Classes
        classes = set(entity.classes)
        labels = _pp_list_of_labels_or_uids(classes, namespace=True)
        pp["classes"] = f"\n  type{'s' if len(classes) > 1 else ''}: {labels}"
        # Attribute values
        values_str = _pp_individual_values(entity)
        if values_str:
            pp["attributes"] = f"\n  values: {values_str}"
        # Subelements
        pp["subelements"] = _pp_individual_subelements(entity, rel)

    pp["identifier"] = f"\n  identifier: {entity.uid}"

    pp = reduce(add, pp.values())
    print(pp, file=file)


def _pp_entity_name(entity: OntologyEntity, type_: bool = False):
    """Return the name of the given element following the pretty print format.

    Args:
        entity: element to be printed.

    Returns:
        String with the pretty printed text.
    """
    type_names = {
        OntologyIndividual: "Ontology individual",
        OntologyClass: "Ontology class",
        OntologyRelationship: "Ontology relationship",
        OntologyAttribute: "Ontology attribute",
        Composition: "Composition",
        Restriction: "Restriction",
    }
    type_name = next(
        (
            name
            for type_, name in type_names.items()
            if isinstance(entity, type_)
        ),
        "Ontology entity",
    )

    title = "- %s" % type_name

    if type_ is True and isinstance(entity, OntologyIndividual):
        classes = entity.classes
        title += (
            f" of class{'es' if len(classes) > 1 else ''} "
            f"{','.join(str(x) for x in classes)}"
        )

    label = entity.label
    if label is not None:
        title += f" named {label}"
    # else:
    #     title += f" {entity.uid}"

    return title


def _pp_individual_subelements(
    individual: OntologyIndividual,
    rel: Union[
        OntologyRelationship, Iterable[OntologyRelationship]
    ] = owl.topObjectProperty,
    level_indentation: str = "\n  ",
    visited: Optional[Set] = None,
) -> str:
    """Recursively formats the subelements from an individual.

    The objects are grouped by ontology class.

    Args:
        individual: element to inspect.
        level_indentation: common characters to left-pad the text.

    Returns:
        String with the pretty printed text
    """
    pp_sub = ""

    # Find relationships connecting the individual to its subelements.
    consider_relationships = set()
    if isinstance(rel, OntologyRelationship):
        rels = {rel}
    else:
        rels = rel
    for predicate in individual.session.graph.predicates(
        individual.identifier, None
    ):
        try:
            relationship = individual.session.ontology.from_identifier(
                predicate
            )
            if isinstance(relationship, OntologyRelationship):
                consider_relationships |= {relationship}
        except KeyError:
            pass
    filtered_relationships = filter(
        lambda x: any(x.is_subclass_of(r) for r in rels),
        consider_relationships,
    )
    sorted_relationships = sorted(filtered_relationships, key=str)

    visited = visited or set()
    visited.add(individual.uid)
    for i, relationship in enumerate(sorted_relationships):
        relationship_name = _pp_list_of_labels_or_uids({relationship})
        pp_sub += level_indentation + (
            " |_Relationship %s (%s):"
            % (relationship_name, relationship.namespace.name)
        )
        sorted_elements = sorted(
            individual.iter(rel=relationship, return_rel=True),
            key=lambda x: (
                "\0"
                + str(
                    sorted(
                        class_.label
                        if class_.label is not None
                        else class_.identifier
                        for class_ in x[0].classes
                    )[0]
                ),
                f"\0{x[1].label}"
                if x[1].label is not None
                else f"{x[1].identifier}",
                f"\0{x[0].label}" if x[0].label is not None else f"{x[0].uid}",
            ),
        )
        for j, (element, rel) in enumerate(sorted_elements):
            if rel != relationship:
                continue
            if i == len(sorted_relationships) - 1:
                indentation = level_indentation + "   "
            else:
                indentation = level_indentation + " | "
            pp_sub += indentation + _pp_entity_name(element, type_=True)
            if j == len(sorted_elements) - 1:
                indentation += "   "
            else:
                indentation += " . "

            identifier_str = f"identifier: {element.uid}"
            pp_sub += indentation + identifier_str

            if element.uid in visited:
                pp_sub += indentation + "(already printed)"
                continue

            values_str = _pp_individual_values(element, indentation)
            if values_str:
                pp_sub += indentation + values_str

            pp_sub += _pp_individual_subelements(
                element, rels, indentation, visited
            )
    return pp_sub


def _pp_individual_values(cuds_object, indentation="\n          "):
    r"""Print the attributes of a cuds object.

    Args:
        cuds_object (Cuds): Print the values of this cuds object.
        indentation (str): The indentation to prepend, defaults to
            "\n          "

    Returns:
        str: The resulting string to print.
    """
    result = []
    sorted_attributes = sorted(
        cuds_object.attributes.items(),
        key=lambda x: (
            f"\0{x[0].label}"
            if x[0].label is not None
            else f"{x[0].identifier}",
            str(x[1]),
        ),
    )
    for attribute, value in sorted_attributes:
        result.append(
            "%s: %s"
            % (
                f"\0{attribute.label}"
                if attribute.label is not None
                else f"{attribute.identifier}",
                value if not len(value) == 1 else next(iter(value)),
            )
        )
    if result:
        return indentation.join(result)


def _pp_list_of_labels_or_uids(
    entities: Iterable[OntologyEntity], namespace: bool = False
) -> str:
    entities = set(entities)
    if namespace:
        labels = (
            f"{entity.label} ({entity.namespace.name})" for entity in entities
        )
    else:
        labels = (entity.label for entity in entities)
    labels = (
        label if label is not None else str(entity.uid)
        for label, entity in zip(labels, entities)
    )
    labels = ", ".join(labels)
    return labels



"""Special kind of ontology individual designed to organize entities."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Iterable, Iterator, Optional, Tuple, Union
from weakref import ReferenceType, ref

from rdflib import URIRef

from simphony_osp.ontology.operations.operations import Operations
from simphony_osp.session.session import Environment, Session
from simphony_osp.utils.simphony_namespace import simphony_namespace

if TYPE_CHECKING:
    from simphony_osp.ontology.individual import OntologyIndividual

logger = logging.getLogger(__name__)


class Container(Operations):
    """Operations associated with the container class.

    Such class and operations are designed to provide a framework that
    allows to cluster several entities together.
    """

    iri: str = str(simphony_namespace.Container)

    _environment: ContainerEnvironment

    def __init__(self, individual: OntologyIndividual):
        """Initialize the container operations."""
        self._environment = ContainerEnvironment(individual=individual)
        super().__init__(individual)

    @property
    def opens_in(self) -> Optional[Union[Environment, URIRef]]:
        """Returns where the container opens by default.

        Returns the environment in which the container opens by
        default.

        This is NOT the session where the container is stored. The
        ontological data defining the container can be stored in ANY
        session, just like any other ontology individual.
        """
        return self._environment.opens_in

    @opens_in.setter
    def opens_in(
        self, value: Optional[Union[OntologyIndividual, Environment]] = None
    ) -> None:
        """Sets where the container opens by default.

        Sets the session or container in which the container opens by
        default.
        """
        from simphony_osp.ontology.individual import OntologyIndividual

        if isinstance(value, OntologyIndividual):
            value = value.operations.environment
        self._environment.opens_in = value

    @property
    def session_linked(self) -> Optional[Session]:
        """Returns session that the container uses to fetch items from."""
        return self._environment.session_linked

    @property
    def is_open(self) -> bool:
        """Returns the current status of the container."""
        return self._environment.is_open

    @property
    def references(self) -> Tuple[URIRef, ...]:
        """Returns a list of the IRIs of all linked individuals.

        - This works without opening the container (checks the linked
          individuals).
        """
        return self._environment.references

    @references.setter
    def references(self, value: Iterable[URIRef]):
        """Replace the current linked individuals with the ones received.

        - This works without opening the container.
        """
        self._environment.references = value

    @property
    def num_references(self) -> int:
        """Returns the number of connected individuals.

        - Does NOT need the container to be open to operate.
        """
        return self._environment.num_references

    def open(self, environment: Optional[Union[Environment]] = None) -> None:
        """Opens the container in `environment`.

        Relevant when an individual needs to be fetched, added
        or removed.

        - PREVENT OPENING CONTAINER IN ITSELF.

        - DO NOT OPEN IF ALREADY OPEN.

        - If no environment (session or container) is provided:
            - Try to open in `opens_in` if not None.
            - Otherwise, try to open in the current container context (if
              exists, the last element of the context list). Beware that the
              `linked_session` of the container context needs to match the
              current default session. Otherwise we changed the default
              session, and we need to open in the default session instead.
            - Otherwise, try to open in the default session (Cuds._session).

        - If opening in a session:
            - Fails (raise RuntimeError) if such session is closed.
            - Changes `_session_linked` to the provided session.
            - Pushes self to the context stack.
        - If opening in a container:
            - If not open, open such container without arguments.
            - Change `_session_linked` to such container's `session_linked`
               attribute.
            - Push self to the context stack.

        - Only really safe in a single threaded program.
        - What happens if the user closes the session within a container?
        """
        return self._environment.open(environment)

    def close(self):
        """Close a container.

        No objects can be added to, removed from or fetched from a
        container while it is closed.

        - Raise RuntimeError if self is not on the top of the stack.
        - Changes `_session_linked` to None.
        - Removes self from the top of the stack.
        """
        return self._environment.close()

    def add(self, *individuals: OntologyIndividual):
        """Adds ontology individuals to a container.

        - Needs the container to be open to operate, tries to open it.
          Closes it after the operation is finished.

        - (Add individual) Adds the individual to
          `session_linked`'s bag if not already there. The
          individual's session may be different from
          `session_linked'. Do the usual stuff in such case
          (put individual in added/updated buffer).

        - (Link individual) Add triples (URIRef(f'<self.iri>'),
          <simphony:contains>,  URIRef(f'{individual.iri}')) for each
          individual in `individuals` to `self.session`'s graph.

          Such a triple means that the individual is linked to the container.

        - Does NOT add duplicate triples to the container (useless), not
          what users would expect.
        """
        return self._environment.add(*individuals)

    def remove(self, *individuals: OntologyIndividual):
        """Removes ontology individuals from a container.

        - Needs the container to be open to operate, tries to open it. Closes
          it after the operation is finished.

        - Raises ValueError if the individual is not linked to the container.

        - (Remove individual) Removes the individual from
          `session_linked`'s bag.

        - Raises ValueError if the individual is not in
          `session_linked`'s bag.

        - (Unlink individual) Remove all triples with the pattern
          (URIRef(f'<self.iri>'), <simphony:contains>,
          URIRef(f'{individual.iri}')) from `self.session`'s graph.  This means
          that the individual is no longer linked to the container.
        """
        return self._environment.remove(*individuals)

    @property
    def members(self) -> Iterator[OntologyIndividual]:
        """Yields ontology individuals from `session_linked`.

        - Needs the container to be open to operate, tries to open it. Closes
          it after the operation is finished.

        - Raise warning when some a linked individual is unreachable in
          `session_linked`.
        """
        return iter(self._environment)

    @property
    def num_members(self) -> int:
        """Returns the number of individuals available to the container.

        - Needs the container to be open to operate, tries to open it. Closes
          it after the operation is finished.

        - Raise warning when some linked individuals are unreachable in
          `session_linked`.
        """
        return len(self._environment)

    def contains(self, entity: OntologyIndividual) -> bool:
        """Checks whether an individual is in the container.

        - Needs the container to be open to operate, tries to open it. Closes
        it after the operation is finished.
        """
        return entity in self._environment

    @property
    def environment(self) -> ContainerEnvironment:
        """Returns the associated container environment."""
        return self._environment

    def connect(self, *references: Iterable[Union[URIRef, str]]):
        """Connects the references to individuals to the container.

        - Does NOT need the container to be open to operate.

        - (Link individual).
        """
        return self._environment.connect(*references)

    def disconnect(self, *references: Union[URIRef, str]):
        """Disconnects the references to individuals from the container.

        - Does NOT need the container to be open to operate.

        - (Unlink individual).
        """
        return self._environment.disconnect(*references)


class ContainerEnvironment(Environment):
    """Environment designed to cluster entities together."""

    _individual: OntologyIndividual

    def __init__(
        self, *args, individual: OntologyIndividual, **kwargs
    ) -> None:
        """Initialize the container."""
        self._individual = individual
        super().__init__(*args, **kwargs)

    @property
    def opens_in(self) -> Optional[Union[Environment, URIRef]]:
        """Returns where the container opens by default.

        Returns the environment in which the container opens by
        default.

        This is NOT the session where the container is stored. The
        ontological data defining the container can be stored in ANY
        session, just like any other ontology individual.
        """
        if isinstance(self._opens_in, ref):
            # Weak reference to environment Python object.
            return self._opens_in()
        else:
            # IRI, try to match to an environment.
            for context in self._stack_default_environment[::-1]:
                if (
                    hasattr(context, "iri")
                    and getattr(context, "iri") == self._opens_in
                ):
                    return context
            else:
                return self._opens_in

    @opens_in.setter
    def opens_in(self, value: Optional[Environment] = None) -> None:
        """Sets where the container opens by default.

        Sets the session or container in which the container opens by
        default.
        """
        if hasattr(value, "iri"):
            self._opens_in = value.iri
        elif isinstance(value, Environment):
            self._opens_in = ref(value)
        elif value is None:
            self._opens_in = value
        else:
            raise TypeError(f"Expected {Environment}, got " f"{type(value)}.")

    _opens_in: Optional[Union[ReferenceType, URIRef]] = None
    """Where the container opens by default.

    Optional[Union[ReferenceType[Environment], URIRef]]
    """

    _container_environment: Optional[Container] = None
    """The container in which this container has been opened (if applies)."""

    @property
    def session_linked(self) -> Optional[Session]:
        """Returns session that the container uses to fetch items from."""
        return self._session_linked

    _session_linked: Optional[Session] = None
    """The session that the container uses to fetch items from."""

    @property
    def is_open(self) -> bool:
        """Returns the current status of the container."""
        return self._session_linked is not None

    # Methods that do NOT require the container to be open.
    # -------------------------------------------------------------------------

    @property
    def references(self) -> Tuple[URIRef, ...]:
        """Returns a list of the IRIs of all linked individuals.

        - This works without opening the container (checks the linked
          individuals).

        """
        return tuple(
            self._individual.session.graph.objects(
                self._individual.identifier, simphony_namespace.contains
            )
        )

    @references.setter
    def references(self, value: Iterable[URIRef]):
        """Replace the current linked individuals with the ones received.

        - This works without opening the container.
        """
        provided = set(value)
        existing = set(
            self._individual.session.graph.objects(
                self._individual.identifier, simphony_namespace.contains
            )
        )
        add = provided - existing
        remove = existing - provided
        for iri in remove:
            self._individual.session.graph.remove(
                (self._individual.identifier, simphony_namespace.contains, iri)
            )
        for iri in add:
            self._individual.session.graph.add(
                (self._individual.identifier, simphony_namespace.contains, iri)
            )

    @property
    def num_references(self) -> int:
        """Returns the number of connected individuals.

        - Does NOT need the container to be open to operate.
        """
        return len(
            set(
                self._individual.session.graph.objects(
                    self._individual.identifier, simphony_namespace.contains
                )
            )
        )

    def open(self, environment: Optional[Union[Environment]] = None) -> None:
        """Opens the container in `environment`.

        Relevant when an individual needs to be fetched, added
        or removed.

        - PREVENT OPENING CONTAINER IN ITSELF.

        - DO NOT OPEN IF ALREADY OPEN.

        - If no environment (session or container) is provided:
            - Try to open in `opens_in` if not None.
            - Otherwise, try to open in the current container context (if
              exists, the last element of the context list). Beware that the
              `linked_session` of the container context needs to match the
              current default session. Otherwise we changed the default
              session, and we need to open in the default session instead.
            - Otherwise, try to open in the default session (Cuds._session).

        - If opening in a session:
            - Fails (raise RuntimeError) if such session is closed.
            - Changes `_session_linked` to the provided session.
            - Pushes self to the context stack.
        - If opening in a container:
            - If not open, open such container without arguments.
            - Change `_session_linked` to such container's `session_linked`
               attribute.
            - Push self to the context stack.

        - Only really safe in a single threaded program.
        - What happens if the user closes the session within a container?
        """
        if environment is None:
            # Try to open in current environment
            environment = self._opens_in or self._stack_default_environment[-1]

        if isinstance(environment, ref):
            # Weak reference to environment Python object.
            environment = environment()
        else:
            # IRI, try to match to an environment.
            for context in self._environment_references:
                if (
                    hasattr(context, "iri")
                    and getattr(context, "iri") == self._opens_in
                ):
                    environment = context
                    break

        if isinstance(environment, Session):
            if False:  # (session_closed) TODO: session should report this.
                raise RuntimeError(
                    f"Cannot open container {self} in closed "
                    f"session {environment}."
                )
            if self.is_open and environment is not self._session_linked:
                raise RuntimeError(
                    f"Attempted to open container {self} in "
                    f"session {environment}, but the "
                    f"has been already opened in "
                    f"{self._session_linked}. Please, "
                    f"close the container first."
                )
            self._session_linked = environment
            self._environment_references.add(self)
            self._session_linked.subscribers.add(self)
        elif isinstance(environment, ContainerEnvironment):
            if (
                self.is_open
                and self._session_linked is not environment._session_linked
            ):
                raise RuntimeError(
                    f"Attempted to open container {self} in "
                    f"container {environment}, which is "
                    f"linked to session "
                    f"{environment._session_linked} "
                    f"different from "
                    f"{self._session_linked} in which this "
                    f"container is already open. Please close "
                    f"the container first."
                )
            if environment is not self:
                if not environment.is_open:
                    environment.open()
                self._container_environment = environment
                self._session_linked = environment._session_linked
                self._container_environment.subscribers.add(self)
                self._environment_references.add(self)
        else:
            raise RuntimeError(
                f"Environment {environment} not found. Try "
                f"opening it if it is a container."
            )

    def close(self):
        """Close a container.

        No objects can be added to, removed from or fetched from a
        container while it is closed.

        - Raise RuntimeError if self is not on the top of the stack.
        - Changes `_session_linked` to None.
        - Removes self from the top of the stack.
        """
        if self in self._stack_default_environment:
            raise RuntimeError(
                "Cannot close a container that is currently "
                "being used as a context manager."
            )
        super().close()
        environment = self._container_environment or self._session_linked
        if environment is not None:
            environment.subscribers.remove(self)
        self._session_linked = None
        self._container_environment = None

    def __enter__(self):
        """Sets the container context to this container.

        Essentially just opens self without arguments.
        """
        self.open()
        Environment.__enter__(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Sets the container context to the previous container.

        Essentially just closes self.
        """
        Environment.__exit__(self, exc_type, exc_val, exc_tb)

    # Methods that require the container to be open.
    # --------------------------------------------------------------

    def add(self, *individuals: OntologyIndividual):
        """Adds ontology individuals to a container.

        - Needs the container to be open to operate, tries to open it.
          Closes it after the operation is finished.

        - (Add individual) Adds the individual to
          `session_linked`'s bag if not already there. The
          individual's session may be different from
          `session_linked'. Do the usual stuff in such case
          (put individual in added/updated buffer).

        - (Link individual) Add triples (URIRef(f'<self.iri>'),
          <simphony:contains>,  URIRef(f'{individual.iri}')) for each
          individual in `individuals` to `self.session`'s graph.

          Such a triple means that the individual is linked to the container.

        - Does NOT add duplicate triples to the container (useless), not
          what users would expect.
        """
        with self:
            for individual in individuals:
                self._session_linked.update(individual)
                self._individual.session.graph.add(
                    (
                        self._individual.identifier,
                        simphony_namespace.contains,
                        individual.identifier,
                    )
                )

            result = list(
                self._session_linked.from_identifier(i.identifier)
                for i in individuals
            )
        return result[0] if len(individuals) == 1 else list(result)

    def remove(self, *individuals: OntologyIndividual):
        """Removes ontology individuals from a container.

        - Needs the container to be open to operate, tries to open it. Closes
          it after the operation is finished.

        - Raises ValueError if the individual is not linked to the container.

        - (Remove individual) Removes the individual from
          `session_linked`'s bag.

        - Raises ValueError if the individual is not in
          `session_linked`'s bag.

        - (Unlink individual) Remove all triples with the pattern
          (URIRef(f'<self.iri>'), <simphony:contains>,
          URIRef(f'{individual.iri}')) from `self.session`'s graph.  This means
          that the individual is no longer linked to the container.
        """
        with self:
            value_errors = set()
            to_remove = set()
            for individual in individuals:
                if individual not in self:
                    value_errors.add(individual)
                else:
                    to_remove.add(individual)

            if value_errors:
                raise ValueError(
                    "The following entities cannot be "
                    "removed, since they are not in "
                    + f"container {self}: %s"
                    % ", ".join(f"{individual}" for individual in value_errors)
                )
            for individual in to_remove:
                self._individual.session.graph.remove(
                    (
                        self._individual.identifier,
                        simphony_namespace.contains,
                        individual.identifier,
                    )
                )

    @property
    def members(self) -> Iterator[OntologyIndividual]:
        """Yields ontology individuals from `session_linked`.

        - Needs the container to be open to operate, tries to open it. Closes
          it after the operation is finished.

        - Raise warning when some a linked individual is unreachable in
          `session_linked`.
        """
        return iter(self)

    @property
    def num_members(self) -> int:
        """Returns the number of individuals available to the container.

        - Needs the container to be open to operate, tries to open it. Closes
          it after the operation is finished.

        - Raise warning when some linked individuals are unreachable in
          `session_linked`.
        """
        return len(self)

    def contains(self, entity: OntologyIndividual) -> bool:
        """Checks whether an individual is in the container.

        - Needs the container to be open to operate, tries to open it. Closes
        it after the operation is finished.
        """
        return entity in self

    @property
    def environment(self) -> ContainerEnvironment:
        """Returns the associated container environment."""
        return self

    def __iter__(self) -> Iterator[OntologyIndividual]:
        """Yields ontology individuals from `session_linked`.

        - Needs the container to be open to operate, tries to open it. Closes
          it after the operation is finished.

        - Raise warning when some a linked individual is unreachable in
          `session_linked`.
        """
        with self:
            for reference in self.references:
                try:
                    yield self.session_linked.from_identifier(reference)
                except KeyError:
                    logger.warning(
                        f"Container {self} is connected to "
                        f"reference {reference}, but such "
                        f"reference is unavailable in its linked "
                        f"session {self._session_linked}."
                    )

    def __len__(self) -> int:
        """Returns the number of individuals available to the container.

        - Needs the container to be open to operate, tries to open it. Closes
          it after the operation is finished.

        - Raise warning when some linked individuals are unreachable in
          `session_linked`.
        """
        length = 0
        with self:
            for reference in self.references:
                try:
                    self.session_linked.from_identifier(reference)
                    length += 1
                except KeyError:
                    pass
        return length

    def __contains__(self, entity: OntologyIndividual) -> bool:
        """Checks whether an individual is in the container.

        - Needs the container to be open to operate, tries to open it. Closes
        it after the operation is finished.
        """
        with self:
            if entity.identifier not in self.references:
                return False
            try:
                self.session_linked.from_identifier(entity.identifier)
                return True
            except KeyError:
                return False

    # Methods that do NOT require the container to be open.
    # -------------------------------------------------------------------------
    def connect(self, *references: Iterable[Union[URIRef, str]]):
        """Connects the references to individuals to the container.

        - Does NOT need the container to be open to operate.

        - (Link individual).
        """
        references = set(map(lambda x: URIRef(x), references))
        self.references = set(self.references).union(references)

    def disconnect(self, *references: Union[URIRef, str]):
        """Disconnects the references to individuals from the container.

        - Does NOT need the container to be open to operate.

        - (Unlink individual).
        """
        references = set(map(lambda x: URIRef(x), references))
        self.references = set(self.references).difference(references)



"""Utilities for searching entities."""

from itertools import chain
from typing import (
    Callable,
    FrozenSet,
    Iterable,
    Iterator,
    Optional,
    Set,
    Union,
)

from rdflib import OWL
from rdflib.term import Node

from simphony_osp.ontology.attribute import OntologyAttribute
from simphony_osp.ontology.individual import OntologyIndividual
from simphony_osp.ontology.oclass import OntologyClass
from simphony_osp.ontology.relationship import OntologyRelationship
from simphony_osp.session.session import QueryResult, Session
from simphony_osp.utils.datatypes import UID, AttributeValue


def find(
    root: OntologyIndividual,
    criterion: Callable[[OntologyIndividual], bool] = (lambda x: True),
    rel: Union[
        Union[OntologyRelationship, Node],
        Iterable[Union[OntologyRelationship, Node]],
    ] = OWL.topObjectProperty,
    find_all: bool = True,
    max_depth: Union[int, float] = float("inf"),
) -> Union[Optional[OntologyIndividual], Iterator[OntologyIndividual]]:
    """Finds a set of ontology individuals following the given relationships.

    Use the given relationship for traversal.

    Args:
        criterion: Function that returns True on the ontology individual that
            is searched.
        root: Starting point of the search.
        rel: The relationship(s) (incl. sub-relationships) to consider for
            traversal.
        find_all: Whether to find all ontology individuals satisfying
            the criterion.
        max_depth: The maximum depth for the search. Defaults to
            float("inf") (unlimited).

    Returns:
        The element(s) found. One element (or `None` is returned when
        `find_all` is `False`, a generator when `find_all` is True.
    """
    if isinstance(rel, (OntologyRelationship, Node)):
        rel = {rel}
    rel = frozenset(rel)

    result = _iter(criterion, root, rel, max_depth)
    if not find_all:
        result = next(result, None)

    return result


def _iter(
    criterion: Callable[[OntologyIndividual], bool],
    root: OntologyIndividual,
    rel: FrozenSet[Union[OntologyRelationship, Node]],
    max_depth: Union[int, float] = float("inf"),
    current_depth: int = 0,
    visited: Optional[Set[UID]] = None,
) -> Iterator[OntologyIndividual]:
    """Finds a set of ontology individuals following the given relationships.

    Use the given relationship for traversal.

    Args:
        criterion: Function that returns True on the ontology individual that
            is searched.
        root: Starting point of the search.
        rel: The relationship(s) (incl. sub-relationships) to consider for
            traversal.
        max_depth: The maximum depth for the search. Defaults to
            float("inf") (unlimited).
        current_depth: The current search depth. Defaults to 0.
        visited: The set of UIDs already visited. Defaults to None.

    Returns:
        The element(s) found.
    """
    visited = visited or set()
    visited.add(root.uid)
    if criterion(root):
        yield root

    if current_depth < max_depth:
        for sub in chain(*(root.iter(rel=r) for r in rel)):
            if sub.uid not in visited:
                yield from _iter(
                    criterion=criterion,
                    root=sub,
                    rel=rel,
                    max_depth=max_depth,
                    current_depth=current_depth + 1,
                    visited=visited,
                )


def find_by_identifier(
    root: OntologyIndividual,
    identifier: Union[Node, UID, str],
    rel: Union[
        Union[OntologyRelationship, Node],
        Iterable[Union[OntologyRelationship, Node]],
    ] = OWL.topObjectProperty,
) -> Optional[OntologyIndividual]:
    """Recursively finds an ontology individual with given identifier.

    Only uses the given relationship for traversal.

    Args:
        root: Starting point of search.
        identifier: The identifier of the entity that is searched.
        rel: The relationship (incl. sub-relationships) to consider.

    Returns:
        The resulting individual.
    """
    return find(
        root=root,
        criterion=lambda individual: individual.uid == UID(identifier),
        rel=rel,
        find_all=False,
    )


def find_by_class(
    root: OntologyIndividual,
    oclass: OntologyClass,
    rel: Union[
        Union[OntologyRelationship, Node],
        Iterable[Union[OntologyRelationship, Node]],
    ] = OWL.topObjectProperty,
) -> Iterator[OntologyIndividual]:
    """Recursively finds ontology individuals with given class.

    Only uses the given relationship for traversal.

    Args:
        root: Starting point of search.
        oclass: The ontology class of the entity that is searched.
        rel: The relationship (incl. sub-relationships) to consider for
            traversal.

    Returns:
        The individuals found.
    """
    return find(
        criterion=lambda individual: individual.is_a(oclass),
        root=root,
        rel=rel,
        find_all=True,
    )


def find_by_attribute(
    root: OntologyIndividual,
    attribute: OntologyAttribute,
    value: AttributeValue,
    rel: Union[
        Union[OntologyRelationship, Node],
        Iterable[Union[OntologyRelationship, Node]],
    ] = OWL.topObjectProperty,
) -> Iterator[OntologyIndividual]:
    """Recursively finds ontology individuals by attribute and value.

    Only the given relationship will be used for traversal.

    Args:
        root: The root for the search.
        attribute: The attribute to look for.
        value: The corresponding value to filter by.
        rel: The relationship (incl. sub-relationships) to consider.

    Returns:
        The individuals found.
    """
    return find(
        criterion=(lambda individual: value in individual[attribute]),
        root=root,
        rel=rel,
        find_all=True,
    )


def find_relationships(
    root: OntologyIndividual,
    find_rel: OntologyRelationship,
    find_sub_relationships: bool = False,
    rel: Union[
        Union[OntologyRelationship, Node],
        Iterable[Union[OntologyRelationship, Node]],
    ] = OWL.topObjectProperty,
) -> Iterator[OntologyIndividual]:
    """Find the given relationship in the subtree of the given root.

    Args:
        root: Only consider the subgraph rooted in this root.
        find_rel: The relationship to find.
        find_sub_relationships: Treat relationships that are a
            sub-relationship of the relationship to find as valid results.
            Defaults to `False`.
        rel: Only consider these relationships (incl. sub-relationships) when
            searching.


    Returns:
        The ontology individuals having the given relationship.
    """
    subclasses = find_rel.subclasses if find_sub_relationships else {find_rel}

    def criterion(individual):
        return any(
            relationship in subclasses
            for _, relationship in individual.relationships_iter(
                return_rel=True
            )
        )

    return find(criterion=criterion, root=root, rel=rel, find_all=True)


def sparql(query: str, session: Optional[Session] = None) -> QueryResult:
    """Performs a SPARQL query on a session.

    Args:
        query: A string with the SPARQL query to perform.
        session: The session on which the SPARQL query will be performed. If no
            session is specified, then the current default session is used.
            This means that, when no session is specified, inside session
            `with` statements, the query will be performed on the session
            associated with such statement, while outside, it will be
            performed on the SimPhoNy default session.

    Returns:
        A QueryResult object, which can be iterated to obtain
        the output rows. Then for each `row`, the value for each query
        variable can be retrieved as follows: `row['variable']`.
    """
    session = session or Session.get_default_session()
    return session.sparql(query)



"""This method contains an RDFLib namespace for the CUBA namespace."""

from pathlib import Path

from rdflib import OWL, RDF, Graph
from rdflib.namespace import ClosedNamespace

file = Path(__file__).parent.parent / "ontology" / "files" / "simphony.ttl"

graph = Graph().parse(file.absolute())

namespace_iri = next(graph.subjects(RDF.type, OWL.Ontology))

entities = (
    IRI[len(namespace_iri) :]
    for type_ in (
        OWL.Class,
        OWL.ObjectProperty,
        OWL.DatatypeProperty,
        OWL.AnnotationProperty,
    )
    for IRI in graph.subjects(RDF.type, type_)
)

simphony_namespace = ClosedNamespace(namespace_iri, entities)
del file, graph, namespace_iri, entities



"""Utilities that do not fit in the other categories."""

from typing import Any, Iterator


def take(iterator: Iterator, amount: int) -> Any:
    """Take a specific number of elements from an iterator.

    Given an iterator, creates another iterator that can yield a maximum of
    `amount` items from the original one.

    Args:
        iterator: Original iterator to take items from.
        amount: Maximum number of items to take from the original iterator.
    """
    for i in range(0, amount):
        try:
            yield next(iterator)
        except StopIteration:
            break



"""Utilities for caching computations."""

from datetime import datetime
from functools import lru_cache, wraps
from typing import Any, Callable, Dict, Optional, Tuple
from weakref import ref

from simphony_osp.utils.other import take


def lru_cache_weak(
    maxsize: Optional[int] = None, typed: bool = False
) -> Callable:
    """Mimics `functools.lru_cache`, but using a weak reference.

    The motivation behind this decorator is that when `lru_cache` is used
    on an instance method, it keeps the references to `self` that are sent
    to it until the cache is cleared, effectively creating a memory leak.

    Using this decorator instead should prevent the memory leak.
    """

    def decorator(func: Callable):
        @lru_cache(maxsize=maxsize, typed=typed)
        def _cached(_self_weak_ref, *args, **kwargs):
            return func(_self_weak_ref(), *args, **kwargs)

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            return _cached(ref(self), *args, **kwargs)

        wrapper.cache_info = _cached.cache_info
        wrapper.cache_clear = _cached.cache_clear
        if hasattr(_cached, "cache_parameters"):
            """The method `cache_parameters` was introduced in Python 3.9."""
            wrapper.cache_parameters = _cached.cache_parameters

        return wrapper

    return decorator


def lru_cache_timestamp(
    read_timestamp: Callable[[Any], datetime],
    maxsize: Optional[int] = None,
    typed: bool = False,
) -> Callable:
    """Like `lru_cache_weak`, but with a timestamp-based cache invalidation.

    The purpose of this decorator is to have a lru-cache decorator that can be
    applied to instance methods whose result depend on a separate object.
    The idea is that when the separate object changes, it can update a
    timestamp stored in itself so that all the cached methods know that they
    need to invalidate the contents of their cache.

    An example of this, and its initial motivation, are ontology classes and
    sessions. The computation of the superclasses of an ontology class
    object is done based on the contents of the graph of the session it is
    attached to. If the session's graph is updated, for example because an
    ontology is installed with pico, the cache should be invalidated. This
    is achieved by updating a timestamp on the session, so that the
    `superclasses` method of each ontology class can compare its own
    timestamp with the session's timestamp, and in this way know that they
    should invalidate their cache.

    Remark: Note that, unlike lru_cache_weak, this decorator keeps an
    independent cache for each value of `self` (that is, for each instance).
    This decision was taken having in mind that each instance can be
    connected to a different external data structure, so having a single
    `lru_cache` would involve that when one of the data structures is modified,
    all the caches are cleared, rather than just the ones related to such data
    structure. Please note that also because of this reason, the real
    maximum size of this cache is `maxsize**2` (`maxsize` instances,
    each one of maximum size `maxsize`). Taking care of this detail was not
    really needed for the use case, so the task was left out.

    Args:
        read_timestamp: A function that retrieves the external timestamp,
            that is, the timestamp of the last cache-invalidating event.
        maxsize: The maximum size of the cache. See the docstring of
            `functools.lru_cache` for more details.
        typed: Whether to cache arguments of different types separately. See
            the docstring of `functools.lru_cache` for more details.
    """

    def decorator(func: Callable):

        holder: Dict[Any, Tuple[datetime, Callable]] = dict()
        """For each instance, holds its timestamp and the cached function."""

        @wraps(func)
        def wrapper(self, *args, **kwargs):

            internal_timestamp, cached_function = holder.get(
                self, (None, None)
            )

            if cached_function is None:
                """Create a cached version of the function for the instance.

                Create a cached version of the function for the given
                instance if none exists yet."""

                internal_timestamp = datetime.now()

                @lru_cache_weak(maxsize=maxsize, typed=typed)
                def cached_function(
                    self_function, *args_function, **kwargs_function
                ):
                    return func(
                        self_function, *args_function, **kwargs_function
                    )

                holder[self] = internal_timestamp, cached_function

            if maxsize is not None:
                # set self as the most recently used item
                try:
                    del holder[self]
                except KeyError:
                    pass
                holder[self] = internal_timestamp, cached_function

                # Remove caches from the holder if there are more instances
                # than the value of the cache size. This is necessary to keep
                # the size of `holder` bounded.
                diff = len(holder) - maxsize
                clear = {item for item in take(holder, diff)}
                for item in clear:
                    del holder[item]

            external_timestamp = read_timestamp(self)
            if external_timestamp and internal_timestamp <= external_timestamp:
                cached_function.cache_clear()
                holder[self] = datetime.now(), cached_function

            return cached_function(self, *args, **kwargs)

        return wrapper

    return decorator



"""Module containing utility functions for developers."""

import simphony_osp.utils.datatypes as datatypes
from simphony_osp.utils.simphony_namespace import simphony_namespace

__all__ = ["datatypes", "simphony_namespace"]
